from typing import Optional, Type

from aiogram.filters import BaseFilter
from aiogram.fsm.state import StatesGroup
from aiogram.types import TelegramObject

from aiogram_dialog.api.entities import Context
from aiogram_dialog.api.internal import CONTEXT_KEY


class IntentFilter(BaseFilter):
    def __init__(self, aiogd_intent_state_group: Optional[Type[StatesGroup]]):
        self.aiogd_intent_state_group = aiogd_intent_state_group

    async def __call__(self, obj: TelegramObject, **kwargs) -> bool:
        del obj  # unused
        if self.aiogd_intent_state_group is None:
            return True

        context: Context = kwargs.get(CONTEXT_KEY)
        if not context:
            return False
        return context.state.group == self.aiogd_intent_state_group



from typing import Callable, Dict, List, Sequence, Tuple, Union

from aiogram_dialog.api.exceptions import InvalidWidgetType
from aiogram_dialog.api.internal import DataGetter
from .data.data_context import CompositeGetter, StaticGetter
from .input import BaseInput, CombinedInput, MessageHandlerFunc, MessageInput
from .kbd import Group, Keyboard
from .media import Media
from .text import Format, Multi, Text
from .widget_event import WidgetEventProcessor

WidgetSrc = Union[str, Text, Keyboard, MessageHandlerFunc, Media, BaseInput]

SingleGetterBase = Union[DataGetter, Dict]
GetterVariant = Union[
    None,
    SingleGetterBase,
    List[SingleGetterBase],
    Tuple[SingleGetterBase, ...],
]


def ensure_text(widget: Union[str, Text, Sequence[Text]]) -> Text:
    if isinstance(widget, str):
        return Format(widget)
    if isinstance(widget, Sequence):
        if len(widget) == 1:
            return widget[0]
        return Multi(*widget)
    return widget


def ensure_keyboard(widget: Union[Keyboard, Sequence[Keyboard]]) -> Keyboard:
    if isinstance(widget, Sequence):
        if len(widget) == 1:
            return widget[0]
        return Group(*widget)
    return widget


def ensure_input(
        widget: Union[
            MessageHandlerFunc,
            WidgetEventProcessor,
            BaseInput,
            Sequence[BaseInput],
        ],
) -> BaseInput:
    if isinstance(widget, BaseInput):
        return widget
    elif isinstance(widget, Sequence):
        if len(widget) == 0:
            return MessageInput(None)
        elif len(widget) == 1:
            return widget[0]
        else:
            return CombinedInput(*widget)
    else:
        return MessageInput(widget)


def ensure_media(widget: Union[Media, Sequence[Media]]) -> Media:
    if isinstance(widget, Media):
        return widget
    if len(widget) > 1:  # TODO case selection of media
        raise ValueError("Only one media widget is supported")
    if len(widget) == 1:
        return widget[0]
    return Media()


def ensure_widgets(
        widgets: Sequence[WidgetSrc],
) -> Tuple[Text, Keyboard, BaseInput, Media]:
    texts = []
    keyboards = []
    inputs = []
    media = []

    for w in widgets:
        if isinstance(w, (str, Text)):
            texts.append(ensure_text(w))
        elif isinstance(w, Keyboard):
            keyboards.append(ensure_keyboard(w))
        elif isinstance(w, (BaseInput, Callable)):
            inputs.append(ensure_input(w))
        elif isinstance(w, Media):
            media.append(ensure_media(w))
        else:
            raise InvalidWidgetType(
                f"Cannot add widget of type {type(w)}. "
                f"Only str, Text, Keyboard, BaseInput "
                f"and Callable are supported",
            )
    return (
        ensure_text(texts),
        ensure_keyboard(keyboards),
        ensure_input(inputs),
        ensure_media(media),
    )


def ensure_data_getter(getter: GetterVariant) -> DataGetter:
    if isinstance(getter, Callable):
        return getter
    elif isinstance(getter, dict):
        return StaticGetter(getter)
    elif isinstance(getter, (list, tuple)):
        return CompositeGetter(*map(ensure_data_getter, getter))
    elif getter is None:
        return StaticGetter({})
    else:
        raise InvalidWidgetType(
            f"Cannot add data getter of type {type(getter)}. "
            f"Only Dict, Callable or List of Callables are supported",
        )



from abc import abstractmethod
from typing import Dict, Optional, Union

from aiogram_dialog.api.internal import TextWidget
from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import (
    BaseWidget, true_condition, Whenable, WhenCondition,
)


class Text(Whenable, BaseWidget, TextWidget):
    def __init__(self, when: WhenCondition = None):
        super().__init__(when=when)

    async def render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        """
        Create text.

        When inheriting override `_render_text` method instead
        if you want to keep processing of `when` condition
        """
        if not self.is_(data, manager):
            return ""
        return await self._render_text(data, manager)

    @abstractmethod
    async def _render_text(self, data, manager: DialogManager) -> str:
        """
        Create text.

        Called if widget is not hidden only (regarding `when`-condition)
        """
        raise NotImplementedError

    def __add__(self, other: Union["Text", str]):
        if isinstance(other, str):
            other = Const(other)
        return Multi(self, other, sep="")

    def __radd__(self, other: Union["Text", str]):
        if isinstance(other, str):
            other = Const(other)
        return Multi(other, self, sep="")

    def find(self, widget_id: str) -> Optional["Text"]:
        # no reimplementation, just change return type
        return super().find(widget_id)


class Const(Text):
    def __init__(self, text: str, when: WhenCondition = None):
        super().__init__(when=when)
        self.text = text

    async def _render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        return self.text


class Multi(Text):
    def __init__(self, *texts: Text, sep="\n", when: WhenCondition = None):
        super().__init__(when=when)
        self.texts = texts
        self.sep = sep

    async def _render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        texts = [await t.render_text(data, manager) for t in self.texts]
        return self.sep.join(filter(None, texts))

    def __iadd__(self, other: Union[Text, str]) -> "Multi":
        if isinstance(other, str):
            other = Const(other)
        self.texts += (other,)
        return self

    def __add__(self, other: Union[Text, str]) -> "Multi":
        if isinstance(other, str):
            other = Const(other)
        if self.condition is true_condition and self.sep == "":
            # reduce nesting
            return Multi(*self.texts, other, sep="")
        else:
            return Multi(self, other, sep="")

    def __radd__(self, other: Union[Text, str]) -> "Multi":
        if isinstance(other, str):
            other = Const(other)
        if self.condition is true_condition and self.sep == "":
            # reduce nesting
            return Multi(other, *self.texts, sep="")
        else:
            return Multi(other, self, sep="")

    def find(self, widget_id: str) -> Optional[Text]:
        for text in self.texts:
            if found := text.find(widget_id):
                return found
        return None



__all__ = [
    "Context", "Data",
    "ChatEvent",
    "LaunchMode",
    "MediaAttachment", "MediaId",
    "ShowMode", "StartMode",
    "MarkupVariant", "NewMessage", "OldMessage", "UnknownText",
    "DEFAULT_STACK_ID", "Stack",
    "DIALOG_EVENT_NAME", "DialogAction", "DialogUpdateEvent",
    "DialogStartEvent", "DialogSwitchEvent", "DialogUpdate",
]

from .context import Context, Data
from .events import ChatEvent
from .launch_mode import LaunchMode
from .media import MediaAttachment, MediaId
from .modes import ShowMode, StartMode
from .new_message import MarkupVariant, NewMessage, OldMessage, UnknownText
from .stack import DEFAULT_STACK_ID, Stack
from .update_event import (
    DIALOG_EVENT_NAME, DialogAction, DialogStartEvent, DialogSwitchEvent,
    DialogUpdate, DialogUpdateEvent,
)



from abc import ABC
from enum import Enum
from typing import Dict, TypedDict, Union

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.internal import RawKeyboard
from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import ManagedScroll, Scroll, WhenCondition
from aiogram_dialog.widgets.text import Const, Format, Text
from .base import Keyboard


class PageDirection(Enum):
    NEXT = "NEXT"
    PREV = "PREV"
    FIRST = "FIRST"
    LAST = "LAST"
    IGNORE = "IGNORE"


class PagerData(TypedDict):
    data: Dict
    current_page: int
    current_page1: int
    pages: int


class PagerPageData(PagerData):
    target_page: int
    target_page1: int


DEFAULT_PAGER_ID = "__pager__"

DEFAULT_LAST_BUTTON_TEXT = Const(">>")
DEFAULT_FIRST_BUTTON_TEXT = Const("<<")
DEFAULT_PREV_BUTTON_TEXT = Const("<")
DEFAULT_NEXT_BUTTON_TEXT = Const(">")
DEFAULT_CURRENT_BUTTON_TEXT = Format("{current_page1}")
DEFAULT_PAGE_TEXT = Format("{target_page1}")
DEFAULT_CURRENT_PAGE_TEXT = Format("[ {current_page1} ]")


class BasePager(Keyboard, ABC):
    def __init__(
            self,
            scroll: Union[str, Scroll, None],
            id: str,
            when: WhenCondition = None,
    ):
        super().__init__(id=id, when=when)
        if isinstance(scroll, str):
            self._scroll_id = scroll
            self._scroll = None
        else:
            self._scroll = scroll
            self._scroll_id = None

    def _find_scroll(self, manager: DialogManager) -> ManagedScroll:
        if self._scroll:
            return self._scroll.managed(manager)
        else:
            return manager.find(self._scroll_id)

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        scroll = self._find_scroll(manager)
        await scroll.set_page(int(data))
        return True


class SwitchPage(BasePager):
    def __init__(
            self, page: Union[int, PageDirection],
            scroll: Union[str, Scroll, None],
            id: str,
            text: Text,
            when: WhenCondition = None,
    ):

        super().__init__(id=id, scroll=scroll, when=when)
        self.page = page
        self.text = text

    async def _get_target_page(
            self, current_page: int, pages: int,
    ) -> int:
        if isinstance(self.page, int):
            return self.page
        if self.page is PageDirection.FIRST:
            return 0

        last_page = pages - 1
        if self.page is PageDirection.PREV:
            return max(0, current_page - 1)
        elif self.page is PageDirection.NEXT:
            return min(last_page, current_page + 1)
        elif self.page is PageDirection.LAST:
            return max(0, last_page)
        else:
            return min(last_page, current_page)

    async def _prepare_data(
            self, data: Dict,
            target_page: int,
            current_page: int, pages: int,
    ) -> PagerPageData:

        return {
            "data": data,
            "target_page": target_page,
            "target_page1": target_page + 1,
            "current_page": current_page,
            "current_page1": current_page + 1,
            "pages": pages,
        }

    async def render_keyboard(
            self, data: Dict, manager: DialogManager,
    ) -> RawKeyboard:
        scroll = self._find_scroll(manager)
        pages = await scroll.get_page_count(data)
        current_page = await scroll.get_page()
        target_page = await self._get_target_page(current_page, pages)
        button_data = await self._prepare_data(
            data=data, target_page=target_page, current_page=current_page,
            pages=pages,
        )
        return await super().render_keyboard(button_data, manager)

    async def _render_keyboard(
            self, data: PagerPageData, manager: DialogManager,
    ) -> RawKeyboard:
        return [[
            InlineKeyboardButton(
                text=await self.text.render_text(data, manager),
                callback_data=self._item_callback_data(data["target_page"]),
            ),
        ]]


class LastPage(SwitchPage):
    def __init__(
            self,
            scroll: Union[str, Scroll, None],
            id: str = DEFAULT_PAGER_ID,
            text: Text = DEFAULT_LAST_BUTTON_TEXT,
            when: WhenCondition = None,
    ):
        super().__init__(
            id=id, text=text,
            page=PageDirection.LAST, scroll=scroll, when=when,
        )


class NextPage(SwitchPage):
    def __init__(
            self,
            scroll: Union[str, Scroll, None],
            id: str = DEFAULT_PAGER_ID,
            text: Text = DEFAULT_NEXT_BUTTON_TEXT,
            when: WhenCondition = None,
    ):
        super().__init__(
            id=id, text=text,
            page=PageDirection.NEXT, scroll=scroll, when=when,
        )


class PrevPage(SwitchPage):
    def __init__(
            self,
            scroll: Union[str, Scroll, None],
            id: str = DEFAULT_PAGER_ID,
            text: Text = DEFAULT_PREV_BUTTON_TEXT,
            when: WhenCondition = None,
    ):
        super().__init__(
            id=id, text=text,
            page=PageDirection.PREV, scroll=scroll, when=when,
        )


class FirstPage(SwitchPage):
    def __init__(
            self,
            scroll: Union[str, Scroll, None],
            id: str = DEFAULT_PAGER_ID,
            text: Text = DEFAULT_FIRST_BUTTON_TEXT,
            when: WhenCondition = None,
    ):
        super().__init__(
            id=id, text=text,
            page=PageDirection.FIRST, scroll=scroll, when=when,
        )


class CurrentPage(SwitchPage):
    def __init__(
            self,
            scroll: Union[str, Scroll, None],
            id: str = DEFAULT_PAGER_ID,
            text: Text = DEFAULT_CURRENT_BUTTON_TEXT,
            when: WhenCondition = None,
    ):
        super().__init__(
            id=id, text=text,
            page=PageDirection.IGNORE, scroll=scroll, when=when,
        )


class NumberedPager(BasePager):
    def __init__(
            self,
            scroll: Union[str, Scroll, None],
            id: str = DEFAULT_PAGER_ID,
            page_text: Text = DEFAULT_PAGE_TEXT,
            current_page_text: Text = DEFAULT_CURRENT_PAGE_TEXT,
            when: WhenCondition = None,
    ):
        super().__init__(id=id, scroll=scroll, when=when)
        self.page_text = page_text
        self.current_page_text = current_page_text

    async def _prepare_data(
            self, data: Dict,
            current_page: int, pages: int,
    ) -> PagerData:
        return {
            "data": data,
            "current_page": current_page,
            "current_page1": current_page + 1,
            "pages": pages,
        }

    async def _prepare_page_data(
            self, data: Dict, target_page: int,
    ) -> PagerData:
        data = data.copy()
        data["target_page"] = target_page
        data["target_page1"] = target_page + 1
        return data

    async def render_keyboard(
            self, data: Dict, manager: DialogManager,
    ) -> RawKeyboard:
        scroll = self._find_scroll(manager)
        pages = await scroll.get_page_count(data)
        current_page = await scroll.get_page()
        pager_data = await self._prepare_data(
            data=data,
            current_page=current_page,
            pages=pages,
        )
        return await super().render_keyboard(pager_data, manager)

    async def _render_keyboard(
            self, data: PagerData, manager: DialogManager,
    ) -> RawKeyboard:
        buttons = []
        pages = data["pages"]
        current_page = data["current_page"]
        for target_page in range(pages):
            button_data = await self._prepare_page_data(
                data=data, target_page=target_page,
            )
            if target_page == current_page:
                text_widget = self.current_page_text
            else:
                text_widget = self.page_text
            text = await text_widget.render_text(button_data, manager)
            buttons.append(InlineKeyboardButton(
                text=text,
                callback_data=self._item_callback_data(target_page),
            ))
        return [buttons]



__all__ = [
    "Const",
    "Text",
    "Format",
    "Multi",
    "Jinja",
    "setup_jinja",
    "List",
    "Case",
    "Progress",
    "ScrollingText",
]

from .base import Const, Multi, Text
from .format import Format
from .jinja import Jinja, setup_jinja
from .list import List
from .multi import Case
from .progress import Progress
from .scrolling_text import ScrollingText



from abc import abstractmethod
from typing import Optional, Protocol

from aiogram import Bot
from aiogram.types import CallbackQuery, Message

from aiogram_dialog.api.entities import NewMessage, OldMessage
from aiogram_dialog.api.exceptions import DialogsError


class MessageNotModified(DialogsError):
    pass


class MessageManagerProtocol(Protocol):
    @abstractmethod
    async def remove_kbd(
            self, bot: Bot, old_message: Optional[Message],
    ) -> Optional[Message]:
        raise NotImplementedError

    @abstractmethod
    async def show_message(
            self, bot: Bot, new_message: NewMessage,
            old_message: Optional[OldMessage],
    ) -> OldMessage:
        raise NotImplementedError

    @abstractmethod
    async def answer_callback(
            self, bot: Bot, callback_query: CallbackQuery,
    ) -> None:
        raise NotImplementedError



from logging import getLogger
from typing import Any, Awaitable, Callable, Dict, Optional

from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import CallbackQuery, Chat, Message, User
from aiogram.types.error_event import ErrorEvent

from aiogram_dialog.api.entities import (
    ChatEvent, Context, DEFAULT_STACK_ID, DialogUpdateEvent, Stack,
)
from aiogram_dialog.api.exceptions import (
    InvalidStackIdError, OutdatedIntent, UnknownIntent, UnknownState,
)
from aiogram_dialog.api.internal import (
    CALLBACK_DATA_KEY, CONTEXT_KEY, STACK_KEY, STORAGE_KEY,
)
from aiogram_dialog.api.protocols import DialogRegistryProtocol
from aiogram_dialog.utils import remove_indent_id
from .storage import StorageProxy

logger = getLogger(__name__)


class IntentMiddlewareFactory:
    def __init__(
            self,
            registry: DialogRegistryProtocol,
    ):
        super().__init__()
        self.registry = registry

    def storage_proxy(self, data: dict):
        proxy = StorageProxy(
            bot=data["bot"],
            storage=data["fsm_storage"],
            user_id=data["event_from_user"].id,
            chat_id=data["event_chat"].id,
            state_groups=self.registry.state_groups(),
        )
        return proxy

    def _check_outdated(self, intent_id: str, stack: Stack):
        """Check if intent id is outdated for stack."""
        if stack.empty():
            raise OutdatedIntent(
                stack.id,
                f"Outdated intent id ({intent_id}) "
                f"for stack ({stack.id})",
            )
        elif intent_id != stack.last_intent_id():
            raise OutdatedIntent(
                stack.id,
                f"Outdated intent id ({intent_id}) "
                f"for stack ({stack.id})",
            )

    async def _load_context(
            self,
            event: ChatEvent,
            intent_id: Optional[str],
            stack_id: Optional[str],
            data: dict,
    ) -> None:
        proxy = self.storage_proxy(data)
        logger.debug(
            "Loading context for intent: `%s`, "
            "stack: `%s`, user: `%s`, chat: `%s`",
            intent_id,
            stack_id,
            event.from_user.id,
            proxy.chat_id,
        )
        if intent_id is not None:
            context = await proxy.load_context(intent_id)
            stack = await proxy.load_stack(context.stack_id)
            self._check_outdated(intent_id, stack)
        elif stack_id is not None:
            stack = await proxy.load_stack(stack_id)
            if stack.empty():
                context = None
            else:
                context = await proxy.load_context(stack.last_intent_id())
        else:
            raise InvalidStackIdError(
                f"Both stack id and intent id are None: {event}",
            )
        data[STORAGE_KEY] = proxy
        data[STACK_KEY] = stack
        data[CONTEXT_KEY] = context

    def _intent_id_from_reply(
            self, event: Message, data: dict,
    ) -> Optional[str]:
        if not (
                event.reply_to_message and
                event.reply_to_message.from_user.id == data["bot"].id and
                event.reply_to_message.reply_markup and
                event.reply_to_message.reply_markup.inline_keyboard
        ):
            return None
        for row in event.reply_to_message.reply_markup.inline_keyboard:
            for button in row:
                if button.callback_data:
                    intent_id, _ = remove_indent_id(button.callback_data)
                    return intent_id
        return None

    async def process_message(
            self,
            handler: Callable,
            event: Message,
            data: dict,
    ):
        if intent_id := self._intent_id_from_reply(event, data):
            await self._load_context(event, intent_id, DEFAULT_STACK_ID, data)
        else:
            await self._load_context(event, None, DEFAULT_STACK_ID, data)
        return await handler(event, data)

    async def process_my_chat_member(
            self,
            handler: Callable,
            event: Message,
            data: dict,
    ) -> None:
        await self._load_context(event, None, DEFAULT_STACK_ID, data)
        return await handler(event, data)

    async def process_aiogd_update(
            self,
            handler: Callable,
            event: DialogUpdateEvent,
            data: dict,
    ):
        await self._load_context(event, event.intent_id, event.stack_id, data)
        return await handler(event, data)

    async def process_callback_query(
            self,
            handler: Callable,
            event: CallbackQuery,
            data: dict,
    ):
        if "event_chat" not in data:
            return await handler(event, data)
        proxy = self.storage_proxy(data)
        data[STORAGE_KEY] = proxy

        original_data = event.data
        if event.data:
            intent_id, callback_data = remove_indent_id(event.data)
            await self._load_context(event, intent_id, DEFAULT_STACK_ID, data)
            data[CALLBACK_DATA_KEY] = original_data
        else:
            await self._load_context(event, None, DEFAULT_STACK_ID, data)
        return await handler(event, data)


SUPPORTED_ERROR_EVENTS = {
    "message",
    "callback_query",
    "my_chat_member",
    "aiogd_update",
}


async def context_saver_middleware(handler, event, data):
    result = await handler(event, data)
    proxy: StorageProxy = data.pop(STORAGE_KEY, None)
    if proxy:
        await proxy.save_context(data.pop(CONTEXT_KEY))
        await proxy.save_stack(data.pop(STACK_KEY))
    return result


class IntentErrorMiddleware(BaseMiddleware):
    def __init__(
            self,
            registry: DialogRegistryProtocol,
    ):
        super().__init__()
        self.registry = registry

    def _is_error_supported(
            self, event: ErrorEvent, data: Dict[str, Any],
    ) -> bool:
        if isinstance(event, InvalidStackIdError):
            return False
        if event.update.event_type not in SUPPORTED_ERROR_EVENTS:
            return False
        if "event_chat" not in data:
            return False
        if "event_from_user" not in data:
            return False
        return True

    async def _fix_broken_stack(
            self, storage: StorageProxy, stack: Stack,
    ) -> None:
        while not stack.empty():
            await storage.remove_context(stack.pop())
        await storage.save_stack(stack)

    async def _load_last_context(
            self, storage: StorageProxy, stack: Stack,
            chat: Chat, user: User,
    ) -> Optional[Context]:
        try:
            return await storage.load_context(stack.last_intent_id())
        except (UnknownIntent, OutdatedIntent):
            logger.warning(
                "Stack is broken for user %s, chat %s, resetting",
                user.id, chat.id,
            )
            await self._fix_broken_stack(storage, stack)
        return None

    async def __call__(
            self,
            handler: Callable[
                [ErrorEvent, Dict[str, Any]], Awaitable[Any],
            ],
            event: ErrorEvent,
            data: Dict[str, Any],
    ) -> Any:
        error = event.exception
        if not self._is_error_supported(event, data):
            return await handler(event, data)

        try:
            chat = data["event_chat"]
            user = data["event_from_user"]

            proxy = StorageProxy(
                bot=data["bot"],
                storage=data["fsm_storage"],
                user_id=user.id,
                chat_id=chat.id,
                state_groups=self.registry.state_groups(),
            )
            data[STORAGE_KEY] = proxy
            if isinstance(error, OutdatedIntent):
                stack = await proxy.load_stack(stack_id=error.stack_id)
            else:
                stack = await proxy.load_stack()
            if stack.empty() or isinstance(error, UnknownState):
                context = None
            else:
                context = await self._load_last_context(
                    storage=proxy, stack=stack, chat=chat, user=user,
                )
            data[STACK_KEY] = stack
            data[CONTEXT_KEY] = context
            return await handler(event, data)
        finally:
            proxy: StorageProxy = data.pop(STORAGE_KEY, None)
            if proxy:
                context = data.pop(CONTEXT_KEY)
                if context is not None:
                    await proxy.save_context(context)
                await proxy.save_stack(data.pop(STACK_KEY))



from abc import abstractmethod
from typing import Any, Dict, Optional, Protocol

from aiogram import Bot
from aiogram.fsm.state import State

from aiogram_dialog.api.entities import (
    ChatEvent, Context, Data, ShowMode, Stack, StartMode,
)


class BaseDialogManager(Protocol):
    @abstractmethod
    async def done(
            self,
            result: Any = None,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def start(
            self,
            state: State,
            data: Data = None,
            mode: StartMode = StartMode.NORMAL,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def switch_to(
            self,
            state: State,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def update(
            self,
            data: Dict,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def bg(
            self,
            user_id: Optional[int] = None,
            chat_id: Optional[int] = None,
            stack_id: Optional[str] = None,
            load: bool = False,  # load chat and user
    ) -> "BaseDialogManager":
        raise NotImplementedError


class BgManagerFactory(Protocol):
    @abstractmethod
    def bg(
            self,
            bot: Bot,
            user_id: int,
            chat_id: int,
            stack_id: Optional[str] = None,
            load: bool = False,  # load chat and user
    ) -> "BaseDialogManager":
        raise NotImplementedError


class DialogManager(BaseDialogManager, Protocol):
    @property
    @abstractmethod
    def event(self) -> ChatEvent:
        raise NotImplementedError

    @abstractmethod
    async def mark_closed(self) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def middleware_data(self) -> Dict:
        """Middleware data."""
        raise NotImplementedError

    @property
    @abstractmethod
    def dialog_data(self) -> Dict:
        """Dialog data for current context."""
        raise NotImplementedError

    @property
    @abstractmethod
    def start_data(self) -> Dict:
        """Start data for current context."""
        raise NotImplementedError

    @property
    @abstractmethod
    def show_mode(self) -> ShowMode:
        """Get current show mode, used for next show action."""
        raise NotImplementedError

    @show_mode.setter
    @abstractmethod
    def show_mode(self, show_mode: ShowMode) -> None:
        """Set current show mode, used for next show action."""
        raise NotImplementedError

    @abstractmethod
    def is_preview(self) -> bool:
        """Check if this manager is used only to generate dialog preview."""
        raise NotImplementedError

    @abstractmethod
    async def show(self, show_mode: Optional[ShowMode] = None) -> None:
        """Show current state to the user."""
        raise NotImplementedError

    @abstractmethod
    async def answer_callback(self) -> None:
        """Answer to a callback query."""
        raise NotImplementedError

    @abstractmethod
    def current_context(self) -> Context:
        """
        Get current dialog context.

        :raise NoContextError if there is no open dialog
        """
        raise NotImplementedError

    @abstractmethod
    def has_context(self) -> bool:
        """Check if there is current context."""
        raise NotImplementedError

    @abstractmethod
    def current_stack(self) -> Stack:
        """Get current dialog stack."""
        raise NotImplementedError

    @abstractmethod
    async def next(self) -> None:
        """Switch to the next state within current dialog."""
        raise NotImplementedError

    @abstractmethod
    async def back(self) -> None:
        """Switch to the previous state within current dialog."""
        raise NotImplementedError

    @abstractmethod
    def find(self, widget_id) -> Optional[Any]:
        """
        Find a widget in current dialog by its id.

        Returns managed adapter for found widget,
        which does not require to pass manager and has only subset of methods.
        """
        raise NotImplementedError

    @abstractmethod
    async def reset_stack(self, remove_keyboard: bool = True) -> None:
        """
        Reset current stack.

        No callbacks are called, contexts are removed from storage.
        """
        raise NotImplementedError

    @abstractmethod
    async def load_data(self) -> Dict:
        """Load data for current state."""
        raise NotImplementedError

    @abstractmethod
    async def close_manager(self) -> None:
        """Release all resources and disable usage of many methods."""
        raise NotImplementedError



from aiogram import Dispatcher
from aiogram.fsm.state import State, StatesGroup

from aiogram_dialog import Dialog, setup_dialogs, Window
from aiogram_dialog.widgets.text import Format


class MainSG(StatesGroup):
    start = State()


def test_register():
    dialog = Dialog(
        Window(
            Format("stub"),
            state=MainSG.start,
        ),
    )

    dp = Dispatcher()
    dp.include_router(dialog)
    setup_dialogs(dp)


def test_name_state_group():
    dialog = Dialog(
        Window(
            Format("stub"),
            state=MainSG.start,
        ),
    )
    assert dialog.name == "MainSG"


def test_name_explicit():
    dialog = Dialog(
        Window(
            Format("stub"),
            state=MainSG.start,
        ),
        name="FooDialog",
    )
    assert dialog.name == "FooDialog"



from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Dict, Protocol, Union

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.internal import Widget
from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor, WidgetEventProcessor,
)
from .action import Actionable
from .managed import ManagedWidget


class Scroll(Widget, Protocol):
    @abstractmethod
    async def get_page_count(self, data: Dict, manager: DialogManager) -> int:
        raise NotImplementedError

    @abstractmethod
    async def get_page(self, manager: DialogManager) -> int:
        raise NotImplementedError

    @abstractmethod
    async def set_page(
            self, event: ChatEvent, page: int, manager: DialogManager,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def managed(self, manager: DialogManager) -> "ManagedScroll":
        raise NotImplementedError


class ManagedScroll(ManagedWidget[Scroll]):
    async def get_page_count(self, data: Dict) -> int:
        return await self.widget.get_page_count(data, self.manager)

    async def get_page(self) -> int:
        return await self.widget.get_page(self.manager)

    async def set_page(self, page: int) -> None:
        return await self.widget.set_page(
            self.manager.event, page, self.manager,
        )


OnPageChanged = Callable[
    [ChatEvent, ManagedScroll, DialogManager],
    Awaitable,
]
OnPageChangedVariants = Union[OnPageChanged, WidgetEventProcessor, None]


class BaseScroll(Actionable, Scroll, ABC):
    def __init__(
            self,
            id: str,
            on_page_changed: OnPageChangedVariants = None,
    ):
        super().__init__(id=id)
        self.on_page_changed = ensure_event_processor(on_page_changed)

    async def get_page(self, manager: DialogManager) -> int:
        return self.get_widget_data(manager, 0)

    async def set_page(
            self, event: ChatEvent, page: int, manager: DialogManager,
    ) -> None:
        self.set_widget_data(manager, page)
        await self.on_page_changed.process_event(
            event,
            self.managed(manager),
            manager,
        )

    def managed(self, manager: DialogManager) -> ManagedScroll:
        return ManagedScroll(self, manager)



from typing import Any, Optional

from aiogram.fsm.state import State
from aiogram.types import CallbackQuery

from aiogram_dialog.api.entities import ChatEvent, Data, StartMode
from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import WhenCondition
from aiogram_dialog.widgets.text import Const, Text
from aiogram_dialog.widgets.widget_event import WidgetEventProcessor
from .button import Button, OnClick

BACK_TEXT = Const("Back")
NEXT_TEXT = Const("Next")
CANCEL_TEXT = Const("Cancel")


class EventProcessorButton(Button, WidgetEventProcessor):
    async def process_event(
            self,
            event: ChatEvent,
            source: Any,
            manager: DialogManager,
            *args,
            **kwargs,
    ):
        await self._on_click(event, self, manager)

    async def _on_click(
            self, callback: CallbackQuery, button: Button,
            manager: DialogManager,
    ):
        raise NotImplementedError


class SwitchTo(EventProcessorButton):
    def __init__(
            self,
            text: Text,
            id: str,
            state: State,
            on_click: Optional[OnClick] = None,
            when: WhenCondition = None,
    ):
        super().__init__(
            text=text, on_click=self._on_click,
            id=id, when=when,
        )
        self.text = text
        self.user_on_click = on_click
        self.state = state

    async def _on_click(
            self, callback: CallbackQuery, button: Button,
            manager: DialogManager,
    ):
        if self.user_on_click:
            await self.user_on_click(callback, self, manager)
        await manager.switch_to(self.state)


class Next(EventProcessorButton):
    def __init__(
            self,
            text: Text = NEXT_TEXT,
            id: str = "__next__",
            on_click: Optional[OnClick] = None,
            when: WhenCondition = None,
    ):
        super().__init__(
            text=text,
            on_click=self._on_click,
            id=id, when=when,
        )
        self.text = text
        self.user_on_click = on_click

    async def _on_click(
            self, callback: CallbackQuery, button: Button,
            manager: DialogManager,
    ):
        if self.user_on_click:
            await self.user_on_click(callback, self, manager)
        await manager.next()


class Back(EventProcessorButton):
    def __init__(
            self,
            text: Text = BACK_TEXT,
            id: str = "__back__",
            on_click: Optional[OnClick] = None,
            when: WhenCondition = None,
    ):
        super().__init__(
            text=text, on_click=self._on_click,
            id=id, when=when,
        )
        self.text = text
        self.callback_data = id
        self.user_on_click = on_click

    async def _on_click(
            self, callback: CallbackQuery, button: Button,
            manager: DialogManager,
    ):
        if self.user_on_click:
            await self.user_on_click(callback, self, manager)
        await manager.back()


class Cancel(EventProcessorButton):
    def __init__(
            self,
            text: Text = CANCEL_TEXT,
            id: str = "__cancel__",
            result: Any = None,
            on_click: Optional[OnClick] = None,
            when: WhenCondition = None,
    ):
        super().__init__(
            text=text, on_click=self._on_click,
            id=id, when=when,
        )
        self.text = text
        self.result = result
        self.user_on_click = on_click

    async def _on_click(
            self, callback: CallbackQuery, button: Button,
            manager: DialogManager,
    ):
        if self.user_on_click:
            await self.user_on_click(callback, self, manager)
        await manager.done(self.result)


class Start(EventProcessorButton):
    def __init__(
            self,
            text: Text,
            id: str,
            state: State,
            data: Data = None,
            on_click: Optional[OnClick] = None,
            mode: StartMode = StartMode.NORMAL,
            when: WhenCondition = None,
    ):
        super().__init__(
            text=text, on_click=self._on_click,
            id=id, when=when,
        )
        self.text = text
        self.start_data = data
        self.user_on_click = on_click
        self.state = state
        self.mode = mode

    async def _on_click(
            self, callback: CallbackQuery, button: Button,
            manager: DialogManager,
    ):
        if self.user_on_click:
            await self.user_on_click(callback, self, manager)
        await manager.start(self.state, self.start_data, self.mode)



import asyncio
import importlib
import inspect
import os.path
import sys
from concurrent.futures import ProcessPoolExecutor
from tempfile import NamedTemporaryFile

from aiohttp import web

from aiogram_dialog.tools.preview import render_preview_content
from aiogram_dialog.tools.transitions import render_transitions


def removesuffix(s, suffix):
    if s.endswith(suffix):
        return s[:-len(suffix)]
    return s


class Renderer:
    def __init__(self, app_module, dialogs_router):
        self.app_module = app_module
        self.dialogs_router = dialogs_router

    async def _get_router(self):
        app_module = importlib.import_module(self.app_module)
        raw_router = getattr(app_module, self.dialogs_router)
        if inspect.iscoroutinefunction(raw_router):
            router = await raw_router()
        elif inspect.isfunction(raw_router):
            router = raw_router()
        else:
            router = raw_router
        return router

    async def _load_preview(self):
        router = await self._get_router()
        return await render_preview_content(router, simulate_events=True)

    async def _load_transitions(self, path: str):
        router = await self._get_router()
        name = removesuffix(path, ".png")
        render_transitions(router, filename=name)

    def load_preview(self):
        return asyncio.run(self._load_preview())

    def load_transitions(self, path):
        return asyncio.run(self._load_transitions(path))


class Controller:
    def __init__(self, app_module, dialogs_router):
        self.renderer = Renderer(app_module, dialogs_router)

    async def preview(self, _request):
        loop = asyncio.get_event_loop()
        with ProcessPoolExecutor(max_workers=1) as executor:
            text = await loop.run_in_executor(
                executor, self.renderer.load_preview,
            )
        return web.Response(
            text=text,
            headers={"Content-Type": "text/html"},
        )

    async def transitions(self, _request):
        loop = asyncio.get_event_loop()
        with NamedTemporaryFile(suffix=".png") as f:
            with ProcessPoolExecutor(max_workers=1) as executor:
                await loop.run_in_executor(
                    executor, self.renderer.load_transitions, f.name,
                )
            return web.Response(
                body=f.read(),
                headers={"Content-Type": "image/png"},
            )


PORT = 9876
INTRO = f"""
Aiogram Dialog
====================

HTML preview:
http://127.0.0.1:{PORT}/

PNG transitions diagram:
http://127.0.0.1:{PORT}/transitions

======================
"""


def disable_print(*_args, **_kwargs):
    pass


def main():
    path, _, app_spec = sys.argv[1].rpartition(os.path.sep)
    if path:
        sys.path.append(path)
    else:
        sys.path.append(os.curdir)
    app_module, dialogs_router = app_spec.split(":")
    controller = Controller(app_module, dialogs_router)
    routes = web.RouteTableDef()
    routes.get("/transitions")(controller.transitions)
    routes.get("/")(controller.preview)

    app = web.Application()
    app.add_routes(routes)
    print(INTRO)  # noqa: T201
    web.run_app(app, port=PORT, print=disable_print)



from enum import Enum


class ShowMode(Enum):
    AUTO = "auto"
    EDIT = "edit"
    SEND = "send"
    DELETE_AND_SEND = "delete_and_send"
    NO_UPDATE = "no_update"


class StartMode(Enum):
    NORMAL = "NORMAL"
    RESET_STACK = "RESET_STACK"
    NEW_STACK = "NEW_STACK"



from .data_context import (
    CompositeGetter,
    DataGetter,
    PreviewAwareGetter,
    StaticGetter,
)

__all__ = [
    "DataGetter",
    "CompositeGetter",
    "StaticGetter",
    "PreviewAwareGetter",
]



from typing import Optional

from aiogram.types import ReplyKeyboardMarkup

from aiogram_dialog import DialogManager
from aiogram_dialog.api.internal.widgets import (
    MarkupFactory, MarkupVariant, RawKeyboard,
)
from aiogram_dialog.utils import transform_to_reply_keyboard
from aiogram_dialog.widgets.text import Text


class ReplyKeyboardFactory(MarkupFactory):
    def __init__(
            self,
            resize_keyboard: Optional[bool] = None,
            one_time_keyboard: Optional[bool] = None,
            input_field_placeholder: Optional[Text] = None,
            selective: Optional[bool] = None,
    ):
        self.resize_keyboard = resize_keyboard
        self.one_time_keyboard = one_time_keyboard
        self.input_field_placeholder = input_field_placeholder
        self.selective = selective

    async def render_markup(
            self, data: dict, manager: DialogManager, keyboard: RawKeyboard,
    ) -> MarkupVariant:
        if self.input_field_placeholder:
            placeholder = await self.input_field_placeholder.render_text(
                data, manager,
            )
        else:
            placeholder = None
        return ReplyKeyboardMarkup(
            keyboard=transform_to_reply_keyboard(keyboard),
            resize_keyboard=self.resize_keyboard,
            one_time_keyboard=self.one_time_keyboard,
            input_field_placeholder=placeholder,
            selective=self.selective,
        )



from aiogram.types import Chat, User

from aiogram_dialog.api.internal import FakeChat, FakeUser
from aiogram_dialog.utils import is_chat_loaded, is_user_loaded


def test_is_chat_loaded():
    assert is_chat_loaded(Chat(id=1, type="private"))
    assert not is_chat_loaded(FakeChat(id=1, type="private"))


def test_is_user_loaded():
    assert is_user_loaded(User(id=1, is_bot=False, first_name=""))
    assert not is_user_loaded(FakeUser(id=1, is_bot=False, first_name=""))



import importlib.metadata

from aiogram.fsm.state import State, StatesGroup

from aiogram_dialog.dialog import Dialog
from aiogram_dialog.widgets.kbd import Cancel, Keyboard, Start
from aiogram_dialog.widgets.text import Const, Jinja, Text
from aiogram_dialog.window import Window


class AiogramDialogStates(StatesGroup):
    ABOUT = State()


async def metadata_getter(**_kwargs) -> dict:
    metadata = importlib.metadata.metadata("aiogram-dialog")
    urls = [u.split(",", maxsplit=1) for u in metadata.get_all("Project-Url")]
    return {
        "metadata": metadata,
        "urls": urls,
    }


def about_dialog():
    return Dialog(
        Window(
            Jinja(
                "<b><u>{{metadata.Name}}</u></b> by @tishka17\n"
                "\n"
                "{{metadata.Summary}}\n"
                "\n"
                "<b>Version:</b> {{metadata.Version}}\n"
                "<b>Author:</b> {{metadata['Author-email']}}\n"
                "\n"
                "{% for name, url in urls%}"
                "<b>{{name}}:</b> <a href=\"{{url}}\">{{url}}</a>\n"
                "{% endfor %}"
                "",
            ),
            Cancel(Const("Ok")),
            getter=metadata_getter,
            preview_data=metadata_getter,
            state=AiogramDialogStates.ABOUT,
            parse_mode="html",
            disable_web_page_preview=True,
        ),
    )


DEFAULT_ABOUT_BTN_TEXT = Const("🗨️ About aiogram-dialog")


def about_aiogram_dialog_button(
        text: Text = DEFAULT_ABOUT_BTN_TEXT,
) -> Keyboard:
    return Start(
        text=text,
        state=AiogramDialogStates.ABOUT,
        id="__aiogd_about__",
    )



from typing import Any, Dict, Optional, Sequence

from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import (
    BaseScroll, OnPageChangedVariants, WhenCondition,
)
from .base import Text
from ..common.items import get_items_getter, ItemsGetterVariant


class List(Text, BaseScroll):
    def __init__(
            self,
            field: Text,
            items: ItemsGetterVariant,
            sep: str = "\n",
            when: WhenCondition = None,
            id: Optional[str] = None,
            page_size: Optional[int] = None,
            on_page_changed: OnPageChangedVariants = None,
    ):
        Text.__init__(self, when=when)
        BaseScroll.__init__(self, id=id, on_page_changed=on_page_changed)
        self.field = field
        self.sep = sep
        self.items_getter = get_items_getter(items)
        self.page_size = page_size

    async def _render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        items = self.items_getter(data)
        pages = self._get_page_count(items)
        if self.page_size is None:
            current_page = 0
            start = 0
        else:
            current_page = await self.get_page(manager)
            start = current_page * self.page_size
            items = items[start:start + self.page_size]

        texts = [
            await self.field.render_text(
                {
                    "current_page": current_page,
                    "current_page1": current_page + 1,
                    "pages": pages,
                    "data": data,
                    "item": item,
                    "pos": pos + 1,
                    "pos0": pos,
                },
                manager,
            )
            for pos, item in enumerate(items, start)
        ]
        return self.sep.join(filter(None, texts))

    async def get_page_count(self, data: Dict, manager: DialogManager) -> int:
        items = self.items_getter(data)
        return self._get_page_count(items)

    def _get_page_count(self, items: Sequence[Any]) -> int:
        if not items:
            return 0
        if self.page_size is None:
            return 1
        return len(items) // self.page_size + bool(len(items) % self.page_size)



import uuid
from datetime import datetime
from typing import Optional

from aiogram import Bot, Dispatcher
from aiogram.types import (
    CallbackQuery, Chat, InlineKeyboardButton, Message, Update, User,
)

from .keyboard import InlineButtonLocator


class FakeBot(Bot):
    def __init__(self):
        pass  # do not call super, so it is invalid bot, used only as a stub

    @property
    def id(self):
        return 1

    def __call__(self, *args, **kwargs) -> None:
        raise RuntimeError("Fake bot should not be used to call telegram")

    def __hash__(self) -> int:
        return 1

    def __eq__(self, other) -> bool:
        return self is other


class BotClient:
    def __init__(
            self,
            dp: Dispatcher,
            user_id: int = 1,
            chat_id: int = 1,
            chat_type: str = "private",
            bot: Optional[Bot] = None,
    ):
        self.chat = Chat(id=chat_id, type=chat_type)
        self.user = User(
            id=user_id, is_bot=False,
            first_name=f"User_{user_id}",
        )
        self.dp = dp
        self.last_update_id = 1
        self.last_message_id = 1
        self.bot = bot or FakeBot()

    def _new_update_id(self):
        self.last_update_id += 1
        return self.last_update_id

    def _new_message_id(self):
        self.last_message_id += 1
        return self.last_message_id

    def _new_message(
            self, text: str, reply_to: Optional[Message],
    ):
        return Message(
            message_id=self._new_message_id(),
            date=datetime.fromtimestamp(1234567890),
            chat=self.chat,
            from_user=self.user,
            text=text,
            reply_to_message=reply_to,
        )

    async def send(self, text: str, reply_to: Optional[Message] = None):
        return await self.dp.feed_update(self.bot, Update(
            update_id=self._new_update_id(),
            message=self._new_message(text, reply_to),
        ))

    def _new_callback(
            self, message: Message, button: InlineKeyboardButton,
    ) -> CallbackQuery:
        if not button.callback_data:
            raise ValueError("Button has no callback data")
        return CallbackQuery(
            id=str(uuid.uuid4()),
            data=button.callback_data,
            chat_instance="--",
            from_user=self.user,
            message=message,
        )

    async def click(
            self, message: Message,
            locator: InlineButtonLocator,
    ) -> str:
        button = locator.find_button(message)
        if not button:
            raise ValueError(
                f"No button matching {locator} found",
            )

        callback = self._new_callback(message, button)
        await self.dp.feed_update(self.bot, Update(
            update_id=self._new_update_id(),
            callback_query=callback,
        ))
        return callback.id



from typing import Any, Callable, Dict, Hashable, Optional, Union

from magic_filter import MagicFilter

from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import WhenCondition
from .base import Text

Selector = Callable[[Dict, "Case", DialogManager], Hashable]


def new_case_field(fieldname: str) -> Selector:
    def case_field(
            data: Dict, widget: "Case", manager: DialogManager,
    ) -> Hashable:
        return data.get(fieldname)

    return case_field


def new_magic_selector(f: MagicFilter) -> Selector:
    def when_magic(
            data: Dict, widget: "Case", manager: DialogManager,
    ) -> bool:
        return f.resolve(data)

    return when_magic


class Case(Text):
    def __init__(
            self,
            texts: Dict[Any, Text],
            selector: Union[str, Selector, MagicFilter],
            when: WhenCondition = None,
    ):
        super().__init__(when=when)
        self.texts = texts
        if isinstance(selector, str):
            self.selector = new_case_field(selector)
        elif isinstance(selector, MagicFilter):
            self.selector = new_magic_selector(selector)
        else:
            self.selector = selector
        self._has_default = ... in self.texts

    async def _render_text(self, data, manager: DialogManager) -> str:
        selection = self.selector(data, self, manager)
        if selection not in self.texts and self._has_default:
            selection = ...
        return await self.texts[selection].render_text(data, manager)

    def find(self, widget_id: str) -> Optional[Text]:
        for text in self.texts.values():
            if found := text.find(widget_id):
                return found
        return None



from typing import Callable, Dict, Union

from magic_filter import MagicFilter

from aiogram_dialog.api.internal import RawKeyboard
from aiogram_dialog.api.protocols import DialogManager
from .base import Keyboard
from ..common.scroll import BaseScroll, OnPageChangedVariants

PagesGetter = Callable[[Dict, "StubScroll", DialogManager], int]


def new_pages_field(fieldname: str) -> PagesGetter:
    def pages_field(
            data: Dict, widget: "StubScroll", manager: DialogManager,
    ) -> int:
        return data.get(fieldname)

    return pages_field


def new_pages_magic(f: MagicFilter) -> PagesGetter:
    def pages_magic(
            data: Dict, widget: "StubScroll", manager: DialogManager,
    ) -> int:
        return f.resolve(data)

    return pages_magic


def new_pages_fixed(pages: int) -> PagesGetter:
    def pages_fixed(
            data: Dict, widget: "StubScroll", manager: DialogManager,
    ) -> int:
        return pages

    return pages_fixed


class StubScroll(Keyboard, BaseScroll):
    def __init__(
            self,
            id: str,
            pages: Union[str, int, PagesGetter, MagicFilter],
            on_page_changed: OnPageChangedVariants = None,
    ):
        Keyboard.__init__(self, id=id, when=None)
        BaseScroll.__init__(self, id=id, on_page_changed=on_page_changed)
        if isinstance(pages, str):
            self._pages = new_pages_field(pages)
        elif isinstance(pages, MagicFilter):
            self._pages = new_pages_magic(pages)
        else:
            self._pages = new_pages_fixed(pages)

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> RawKeyboard:
        return [[]]

    async def get_page_count(self, data: Dict, manager: DialogManager) -> int:
        return self._pages(data, self, manager)



import asyncio
from contextvars import copy_context

from aiogram import Bot, Dispatcher, Router

from aiogram_dialog.api.entities import DialogUpdate


class Updater:
    def __init__(self, dp: Router):
        while dp.parent_router:
            dp = dp.parent_router
        if not isinstance(dp, Dispatcher):
            raise TypeError("Root router must be Dispatcher.")
        self.dp = dp

    async def notify(self, bot: Bot, update: DialogUpdate) -> None:
        def callback():
            asyncio.create_task(
                self._process_update(bot, update),
            )

        asyncio.get_running_loop().call_soon(callback, context=copy_context())

    async def _process_update(self, bot: Bot, update: DialogUpdate) -> None:
        event = update.event
        await self.dp.propagate_event(
            update_type="update",
            event=update,
            bot=bot,
            event_from_user=event.from_user,
            event_chat=event.chat,
            **self.dp.workflow_data,
        )



from abc import abstractmethod
from typing import Any, Dict, List, Optional, Protocol, runtime_checkable, Type

from aiogram.fsm.state import State, StatesGroup

from aiogram_dialog.api.entities import (
    Data, LaunchMode, NewMessage,
)
from .manager import DialogManager


@runtime_checkable
class DialogProtocol(Protocol):
    @property
    def launch_mode(self) -> LaunchMode:
        raise NotImplementedError

    @abstractmethod
    def states_group_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def states(self) -> List[State]:
        raise NotImplementedError

    @abstractmethod
    def states_group(self) -> Type[StatesGroup]:
        raise NotImplementedError

    @abstractmethod
    async def process_close(
            self, result: Any, manager: DialogManager,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def process_start(
            self,
            manager: "DialogManager",
            start_data: Any,
            state: Optional[State] = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def process_result(
            self, start_data: Data, result: Any,
            manager: "DialogManager",
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def find(self, widget_id) -> Any:
        raise NotImplementedError

    @abstractmethod
    async def load_data(
            self, manager: DialogManager,
    ) -> Dict:
        raise NotImplementedError

    @abstractmethod
    async def render(self, manager: DialogManager) -> NewMessage:
        raise NotImplementedError



from logging import getLogger

from aiogram_dialog.api.entities import (
    DialogAction,
    DialogStartEvent,
    DialogSwitchEvent,
    DialogUpdateEvent,
)
from .manager import ManagerImpl
from .. import ShowMode

logger = getLogger(__name__)


async def handle_update(event: DialogUpdateEvent, dialog_manager: ManagerImpl):
    dialog_manager.show_mode = event.show_mode or ShowMode.AUTO
    if isinstance(event, DialogStartEvent):
        await dialog_manager.start(
            state=event.new_state,
            data=event.data,
            mode=event.mode,
            show_mode=event.show_mode,
        )
    elif isinstance(event, DialogSwitchEvent):
        await dialog_manager.switch_to(state=event.new_state)
        await dialog_manager.show()
    elif event.action is DialogAction.UPDATE:
        if not dialog_manager.has_context():
            logger.warning("No context found")
            return
        if event.data:
            for k, v in event.data.items():
                dialog_manager.current_context().dialog_data[k] = v
        await dialog_manager.show()
    elif event.action is DialogAction.DONE:
        await dialog_manager.done(result=event.data)



from itertools import chain
from typing import Dict, Iterable, List, Optional

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.internal import ButtonVariant, RawKeyboard
from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import WhenCondition
from .base import Keyboard


class Group(Keyboard):
    def __init__(
            self,
            *buttons: Keyboard,
            id: Optional[str] = None,
            width: int = None,
            when: WhenCondition = None,
    ):
        super().__init__(id=id, when=when)
        self.buttons = buttons
        self.width = width

    def find(self, widget_id):
        widget = super(Group, self).find(widget_id)
        if widget:
            return widget
        for btn in self.buttons:
            widget = btn.find(widget_id)
            if widget:
                return widget
        return None

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> RawKeyboard:
        kbd: RawKeyboard = []
        for b in self.buttons:
            b_kbd = await b.render_keyboard(data, manager)
            if self.width is None:
                kbd += b_kbd
            else:
                if not kbd:
                    kbd.append([])
                kbd[0].extend(chain.from_iterable(b_kbd))
        if self.width and kbd:
            kbd = self._wrap_kbd(kbd[0])
        return kbd

    def _wrap_kbd(
            self,
            kbd: Iterable[InlineKeyboardButton],
    ) -> RawKeyboard:
        res: RawKeyboard = []
        row: List[ButtonVariant] = []
        for b in kbd:
            row.append(b)
            if len(row) >= self.width:
                res.append(row)
                row = []
        if row:
            res.append(row)
        return res

    async def _process_other_callback(
            self,
            callback: CallbackQuery,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        for b in self.buttons:
            if await b.process_callback(callback, dialog, manager):
                return True
        return False


class Row(Group):
    def __init__(
            self,
            *buttons: Keyboard,
            id: Optional[str] = None,
            when: WhenCondition = None,
    ):
        super().__init__(
            *buttons, id=id, width=9999, when=when,
        )  # telegram doe not allow even 100 columns


class Column(Group):
    def __init__(
            self,
            *buttons: Keyboard,
            id: Optional[str] = None,
            when: WhenCondition = None,
    ):
        super().__init__(*buttons, id=id, when=when, width=1)



import pytest
from aiogram import Dispatcher
from aiogram.filters import CommandStart
from aiogram.fsm.state import State, StatesGroup
from aiogram.types import Message

from aiogram_dialog import (
    Dialog, DialogManager, setup_dialogs, StartMode, Window,
)
from aiogram_dialog.test_tools import BotClient, MockMessageManager
from aiogram_dialog.test_tools.keyboard import InlineButtonTextLocator
from aiogram_dialog.widgets.kbd import Cancel
from aiogram_dialog.widgets.text import Const, Format


class MainSG(StatesGroup):
    start = State()


class SecondarySG(StatesGroup):
    start = State()


class ThirdSG(StatesGroup):
    start = State()


async def start(message: Message, dialog_manager: DialogManager):
    await dialog_manager.start(MainSG.start, mode=StartMode.RESET_STACK)


async def on_start_main(data, dialog_manager: DialogManager):
    await dialog_manager.start(SecondarySG.start)


async def on_start_sub(_, dialog_manager: DialogManager):
    await dialog_manager.start(ThirdSG.start)


async def on_process_result_sub(_, __, dialog_manager: DialogManager):
    await dialog_manager.done()


@pytest.fixture()
def message_manager() -> MockMessageManager:
    return MockMessageManager()


@pytest.fixture()
def client(dp) -> BotClient:
    return BotClient(dp)


@pytest.fixture()
def dp(message_manager: MockMessageManager):
    dp = Dispatcher()
    dp.message.register(start, CommandStart())

    dp.include_router(Dialog(
        Window(
            Const("First"),
            state=MainSG.start,
        ),
        on_start=on_start_main,
    ))
    dp.include_router(Dialog(
        Window(
            Format("Subdialog"),
            Cancel(),
            state=SecondarySG.start,
        ),
        on_process_result=on_process_result_sub,
        on_start=on_start_sub,
    ))
    dp.include_router(Dialog(
        Window(
            Format("Third"),
            Cancel(),
            state=ThirdSG.start,
        ),
    ))
    setup_dialogs(dp, message_manager=message_manager)
    return dp


@pytest.mark.asyncio
async def test_start(dp, message_manager, client):
    # start
    await client.send("/start")
    first_message = message_manager.one_message()
    assert first_message.text == "Third"
    assert first_message.reply_markup

    message_manager.reset_history()
    await client.click(first_message, InlineButtonTextLocator("Cancel"))
    second_message = message_manager.one_message()
    assert second_message.text == "First"
    assert second_message.reply_markup



class DialogsError(RuntimeError):
    pass


# intents and stack
class InvalidStackIdError(DialogsError):
    pass


class UnknownIntent(DialogsError):
    pass


class OutdatedIntent(DialogsError):
    def __init__(self, stack_id, text):
        super().__init__(text)
        self.stack_id = stack_id


class UnknownState(DialogsError):
    pass


class DialogStackOverflow(DialogsError):
    pass


# manager
class IncorrectBackgroundError(DialogsError):
    pass


class NoContextError(DialogsError):
    pass


# navigation
class UnregisteredDialogError(DialogsError):
    pass


class UnregisteredWindowError(DialogsError):
    pass


# widgets
class InvalidWidgetIdError(DialogsError):
    pass


class InvalidWidget(DialogsError):
    pass


class InvalidWidgetType(InvalidWidget):
    pass


class InvalidKeyboardType(DialogsError):
    pass



from __future__ import annotations

from calendar import MONDAY
from dataclasses import dataclass
from datetime import date, datetime, timedelta, timezone
from enum import Enum
from time import mktime
from typing import (
    Any, Callable, Dict, List, Optional, Protocol, TypedDict, Union,
)

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.internal import RawKeyboard
from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import ManagedWidget, WhenCondition
from aiogram_dialog.widgets.text import Format, Text
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor, WidgetEventProcessor,
)
from .base import Keyboard

CALLBACK_NEXT_MONTH = "+"
CALLBACK_PREV_MONTH = "-"
CALLBACK_NEXT_YEAR = "+Y"
CALLBACK_PREV_YEAR = "-Y"
CALLBACK_NEXT_YEARS_PAGE = "+YY"
CALLBACK_PREV_YEARS_PAGE = "-YY"
CALLBACK_SCOPE_MONTHS = "M"
CALLBACK_SCOPE_YEARS = "Y"

CALLBACK_PREFIX_MONTH = "MONTH"
CALLBACK_PREFIX_YEAR = "YEAR"

ZOOM_OUT_TEXT = Format("🔍")

PREV_YEARS_PAGE_TEXT = Format("<< {date:%Y}")
NEXT_YEARS_PAGE_TEXT = Format("{date:%Y} >>")
THIS_YEAR_TEXT = Format("[ {date:%Y} ]")
YEAR_TEXT = Format("{date:%Y}")
PREV_YEAR_TEXT = Format("<< {date:%Y}")
NEXT_YEAR_TEXT = Format("{date:%Y} >>")
MONTHS_HEADER_TEXT = Format("🗓 {date:%Y}")
THIS_MONTH_TEXT = Format("[ {date:%B} ]")
MONTH_TEXT = Format("{date:%B}")
DATE_TEXT = Format("{date:%d}")
TODAY_TEXT = Format("[ {date:%d} ]")
WEEK_DAY_TEXT = Format("{date:%a}")
PREV_MONTH_TEXT = Format("<< {date:%B %Y}")
NEXT_MONTH_TEXT = Format("{date:%B %Y} >>")
DAYS_HEADER_TEXT = Format("🗓 {date:%B %Y}")

BEARING_DATE = date(2018, 1, 1)


def empty_button():
    return InlineKeyboardButton(text=" ", callback_data="")


class CalendarScope(Enum):
    DAYS = "DAYS"
    MONTHS = "MONTHS"
    YEARS = "YEARS"


def month_begin(offset: date):
    return offset.replace(day=1)


def next_month_begin(offset: date):
    return month_begin(month_begin(offset) + timedelta(days=31))


def prev_month_begin(offset: date):
    return month_begin(month_begin(offset) - timedelta(days=1))


def get_today(tz: timezone):
    return datetime.now(tz).date()


class CalendarData(TypedDict):
    current_scope: str
    current_offset: str


class OnDateSelected(Protocol):
    async def __call__(
            self,
            event: ChatEvent,
            widget: ManagedCalendar,
            dialog_manager: DialogManager,
            date: date,
    ) -> Any:
        raise NotImplementedError


@dataclass
class CalendarUserConfig:
    firstweekday: int = MONDAY
    timezone: timezone = datetime.now().astimezone().tzinfo


@dataclass(frozen=True)
class CalendarConfig:
    min_date: date = date(1900, 1, 1)
    max_date: date = date(2100, 12, 31)
    month_columns: int = 3
    years_per_page: int = 20
    years_columns: int = 5


class CalendarScopeView(Protocol):
    """
    Calendar widget part.

    Used to represent keyboard for one of calendar scopes.
    """

    async def render(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        """
        Render keyboard for current scope.

        :param config: configuration related to current user
        :param offset: current offset from which we show dates
        :param data: data received from window getter
        :param manager: dialog manager instance
        :return: rendered keyboard
        """


CallbackGenerator = Callable[[str], str]


class CalendarDaysView(CalendarScopeView):
    def __init__(
            self,
            callback_generator: CallbackGenerator,
            config: CalendarConfig,
            date_text: Text = DATE_TEXT,
            today_text: Text = TODAY_TEXT,
            weekday_text: Text = WEEK_DAY_TEXT,
            header_text: Text = DAYS_HEADER_TEXT,
            zoom_out_text: Text = ZOOM_OUT_TEXT,
            next_month_text: Text = NEXT_MONTH_TEXT,
            prev_month_text: Text = PREV_MONTH_TEXT,
    ):
        self.config = config
        self.zoom_out_text = zoom_out_text
        self.next_month_text = next_month_text
        self.prev_month_text = prev_month_text
        self.callback_generator = callback_generator
        self.date_text = date_text
        self.today_text = today_text
        self.weekday_text = weekday_text
        self.header_text = header_text

    async def _render_date_button(
            self,
            selected_date: date,
            today: date,
            data: Dict,
            manager: DialogManager,
    ) -> InlineKeyboardButton:
        current_data = {
            "date": selected_date,
            "data": data,
        }
        if selected_date == today:
            text = self.today_text
        else:
            text = self.date_text
        raw_date = int(mktime(selected_date.timetuple()))
        return InlineKeyboardButton(
            text=await text.render_text(
                current_data, manager,
            ),
            callback_data=self.callback_generator(str(raw_date)),
        )

    async def _render_days(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        keyboard = []
        # align beginning
        start_date = offset.replace(day=1)  # month beginning
        min_date = max(self.config.min_date, start_date)
        days_since_week_start = start_date.weekday() - config.firstweekday
        if days_since_week_start < 0:
            days_since_week_start += 7
        start_date -= timedelta(days=days_since_week_start)
        end_date = next_month_begin(offset) - timedelta(days=1)
        # align ending
        max_date = min(self.config.max_date, end_date)
        days_since_week_start = end_date.weekday() - config.firstweekday
        days_till_week_end = (6 - days_since_week_start) % 7
        end_date += timedelta(days=days_till_week_end)
        # add days
        today = get_today(config.timezone)
        for offset in range(0, (end_date - start_date).days, 7):
            row = []
            for row_offset in range(7):
                days_offset = timedelta(days=(offset + row_offset))
                current_date = start_date + days_offset
                if min_date <= current_date <= max_date:
                    row.append(await self._render_date_button(
                        current_date, today, data, manager,
                    ))
                else:
                    row.append(empty_button())
            keyboard.append(row)
        return keyboard

    async def _render_week_header(
            self,
            config: CalendarUserConfig,
            data: Dict,
            manager: DialogManager,
    ) -> List[InlineKeyboardButton]:
        week_range = range(config.firstweekday, config.firstweekday + 7)
        header = []
        for week_day in week_range:
            week_day = week_day % 7 + 1
            data = {
                "week_day": week_day,
                "date": BEARING_DATE.replace(day=week_day),
                "data": data,
            }
            header.append(InlineKeyboardButton(
                text=await self.weekday_text.render_text(data, manager),
                callback_data="",
            ))
        return header

    async def _render_pager(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[InlineKeyboardButton]:
        curr_month = offset.month
        next_month = (curr_month % 12) + 1
        prev_month = (curr_month - 2) % 12 + 1
        prev_end = month_begin(offset) - timedelta(1)
        prev_begin = month_begin(prev_end)
        next_begin = next_month_begin(offset)
        if (
                prev_end < self.config.min_date and
                next_begin > self.config.max_date
        ):
            return []

        prev_month_data = {
            "month": prev_month,
            "date": prev_begin,
            "data": data,
        }
        curr_month_data = {
            "month": curr_month,
            "date": BEARING_DATE.replace(month=curr_month),
            "data": data,
        }
        next_month_data = {
            "month": next_month,
            "date": next_begin,
            "data": data,
        }
        if prev_end < self.config.min_date:
            prev_button = empty_button()
        else:
            prev_button = InlineKeyboardButton(
                text=await self.prev_month_text.render_text(
                    prev_month_data, manager,
                ),
                callback_data=self.callback_generator(CALLBACK_PREV_MONTH),
            )
        zoom_button = InlineKeyboardButton(
            text=await self.zoom_out_text.render_text(
                curr_month_data, manager,
            ),
            callback_data=self.callback_generator(CALLBACK_SCOPE_MONTHS),
        )
        if next_begin > self.config.max_date:
            next_button = empty_button()
        else:
            next_button = InlineKeyboardButton(
                text=await self.next_month_text.render_text(
                    next_month_data, manager,
                ),
                callback_data=self.callback_generator(CALLBACK_NEXT_MONTH),
            )

        return [prev_button, zoom_button, next_button]

    async def _render_header(
            self, config, offset, data, manager,
    ) -> List[InlineKeyboardButton]:
        data = {
            "date": offset,
            "data": data,
        }
        return [InlineKeyboardButton(
            text=await self.header_text.render_text(data, manager),
            callback_data=self.callback_generator(CALLBACK_SCOPE_MONTHS),
        )]

    async def render(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        return [
            await self._render_header(config, offset, data, manager),
            await self._render_week_header(config, data, manager),
            *await self._render_days(config, offset, data, manager),
            await self._render_pager(config, offset, data, manager),
        ]


class CalendarMonthView(CalendarScopeView):
    def __init__(
            self,
            callback_generator: CallbackGenerator,
            config: CalendarConfig,
            month_text: Text = MONTH_TEXT,
            this_month_text: Text = THIS_MONTH_TEXT,
            header_text: Text = MONTHS_HEADER_TEXT,
            zoom_out_text: Text = ZOOM_OUT_TEXT,
            next_year_text: Text = NEXT_YEAR_TEXT,
            prev_year_text: Text = PREV_YEAR_TEXT,
    ):
        self.config = config
        self.callback_generator = callback_generator
        self.month_text = month_text
        self.this_month_text = this_month_text
        self.header_text = header_text
        self.zoom_out_text = zoom_out_text
        self.next_year_text = next_year_text
        self.prev_year_text = prev_year_text

    async def _render_pager(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[InlineKeyboardButton]:
        curr_year = offset.year
        next_year = curr_year + 1
        prev_year = curr_year - 1

        if curr_year not in range(
                self.config.min_date.year, self.config.max_date.year,
        ):
            return []

        prev_year_data = {
            "year": prev_year,
            "date": max(
                BEARING_DATE.replace(year=prev_year),
                self.config.min_date,
            ),
            "data": data,
        }
        curr_year_data = {
            "year": curr_year,
            "date": BEARING_DATE.replace(year=curr_year),
            "data": data,
        }
        next_year_data = {
            "year": next_year,
            "date": min(
                BEARING_DATE.replace(year=next_year),
                self.config.max_date,
            ),
            "data": data,
        }
        if prev_year < self.config.min_date.year:
            prev_button = empty_button()
        else:
            prev_button = InlineKeyboardButton(
                text=await self.prev_year_text.render_text(
                    prev_year_data, manager,
                ),
                callback_data=self.callback_generator(CALLBACK_PREV_YEAR),
            )
        if next_year > self.config.max_date.year:
            next_button = empty_button()
        else:
            next_button = InlineKeyboardButton(
                text=await self.next_year_text.render_text(
                    next_year_data, manager,
                ),
                callback_data=self.callback_generator(CALLBACK_NEXT_YEAR),
            )
        zoom_button = InlineKeyboardButton(
            text=await self.zoom_out_text.render_text(
                curr_year_data, manager,
            ),
            callback_data=self.callback_generator(CALLBACK_SCOPE_YEARS),
        )
        return [prev_button, zoom_button, next_button]

    def _is_month_allowed(
            self, offset: date, month: int,
    ) -> bool:
        start = date(offset.year, month, 1)
        end = next_month_begin(start) - timedelta(days=1)
        return (
            end >= self.config.min_date and
            start <= self.config.max_date
        )

    async def _render_month_button(
            self,
            month: int,
            this_month: int,
            data: Dict,
            offset: date,
            config: CalendarUserConfig,
            manager: DialogManager,
    ) -> InlineKeyboardButton:
        if not self._is_month_allowed(offset, month):
            return empty_button()

        month_data = {
            "month": month,
            "date": BEARING_DATE.replace(month=month),
            "data": data,
        }
        if month == this_month:
            text = self.this_month_text
        else:
            text = self.month_text

        return InlineKeyboardButton(
            text=await text.render_text(
                month_data, manager,
            ),
            callback_data=self.callback_generator(
                f"{CALLBACK_PREFIX_MONTH}{month}",
            ),
        )

    async def _render_months(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        keyboard = []
        today = get_today(config.timezone)
        if offset.year == today.year:
            this_month = today.month
        else:
            this_month = -1
        for row in range(1, 13, self.config.month_columns):
            keyboard_row = []
            for column in range(self.config.month_columns):
                month = row + column
                keyboard_row.append(await self._render_month_button(
                    month, this_month, data, offset, config, manager,
                ))
            keyboard.append(keyboard_row)
        return keyboard

    async def _render_header(
            self, config, offset, data, manager,
    ) -> List[InlineKeyboardButton]:
        data = {
            "date": offset,
            "data": data,
        }
        return [InlineKeyboardButton(
            text=await self.header_text.render_text(data, manager),
            callback_data=self.callback_generator(CALLBACK_SCOPE_YEARS),
        )]

    async def render(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        return [
            await self._render_header(config, offset, data, manager),
            *await self._render_months(config, offset, data, manager),
            await self._render_pager(config, offset, data, manager),
        ]


class CalendarYearsView(CalendarScopeView):
    def __init__(
            self,
            callback_generator: CallbackGenerator,
            config: CalendarConfig,
            year_text: Text = YEAR_TEXT,
            this_year_text: Text = THIS_YEAR_TEXT,
            next_page_text: Text = NEXT_YEARS_PAGE_TEXT,
            prev_page_text: Text = PREV_YEARS_PAGE_TEXT,
    ):
        self.callback_generator = callback_generator
        self.config = config
        self.year_text = year_text
        self.this_year_text = this_year_text
        self.next_page_text = next_page_text
        self.prev_page_text = prev_page_text

    async def _render_pager(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[InlineKeyboardButton]:
        curr_year = offset.year
        next_year = curr_year + self.config.years_per_page
        prev_year = curr_year - self.config.years_per_page

        prev_year_data = {
            "year": prev_year,
            "date": max(
                BEARING_DATE.replace(year=prev_year),
                self.config.min_date,
            ),
            "data": data,
        }
        next_year_data = {
            "year": next_year,
            "date": min(
                BEARING_DATE.replace(year=next_year),
                self.config.max_date,
            ),
            "data": data,
        }
        if curr_year <= self.config.min_date.year:
            prev_button = empty_button()
        else:
            prev_button = InlineKeyboardButton(
                text=await self.prev_page_text.render_text(
                    prev_year_data, manager,
                ),
                callback_data=self.callback_generator(
                    CALLBACK_PREV_YEARS_PAGE,
                ),
            )
        if next_year > self.config.max_date.year:
            next_button = empty_button()
        else:
            next_button = InlineKeyboardButton(
                text=await self.next_page_text.render_text(
                    next_year_data, manager,
                ),
                callback_data=self.callback_generator(
                    CALLBACK_NEXT_YEARS_PAGE,
                ),
            )

        if prev_button == next_button == empty_button():
            return []
        return [prev_button, next_button]

    def _is_year_allowed(self, year: int) -> bool:
        return self.config.min_date.year <= year <= self.config.max_date.year

    async def _render_year_button(
            self,
            year: int,
            this_year: int,
            data: Dict,
            config: CalendarUserConfig,
            manager: DialogManager,
    ) -> InlineKeyboardButton:
        if not self._is_year_allowed(year):
            return empty_button()
        if year == this_year:
            text = self.this_year_text
        else:
            text = self.year_text

        year_data = {
            "year": year,
            "date": BEARING_DATE.replace(year=year),
            "data": data,
        }
        return InlineKeyboardButton(
            text=await text.render_text(
                year_data, manager,
            ),
            callback_data=self.callback_generator(
                f"{CALLBACK_PREFIX_YEAR}{year}",
            ),
        )

    async def _render_years(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        keyboard = []
        this_year = get_today(config.timezone).year
        years_columns = self.config.years_columns
        years_per_page = self.config.years_per_page

        for row in range(0, years_per_page, years_columns):
            keyboard_row = []
            for column in range(years_columns):
                curr_year = offset.year + row + column
                keyboard_row.append(await self._render_year_button(
                    curr_year, this_year, data, config, manager,
                ))
            keyboard.append(keyboard_row)
        return keyboard

    async def render(
            self,
            config: CalendarUserConfig,
            offset: date,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        return [
            *await self._render_years(config, offset, data, manager),
            await self._render_pager(config, offset, data, manager),
        ]


class Calendar(Keyboard):
    """
    Calendar widget.

    Used to render keyboard for date selection.
    """

    def __init__(
            self,
            id: str,
            on_click: Union[OnDateSelected, WidgetEventProcessor, None] = None,
            config: Optional[CalendarConfig] = None,
            when: WhenCondition = None,
    ) -> None:
        """
        Init calendar widget.

        :param id: ID of widget
        :param on_click: Function to handle date selection
        :param config: Calendar configuration
        :param when: Condition when to show widget
        """
        super().__init__(id=id, when=when)
        self.on_click = ensure_event_processor(on_click)
        if config is None:
            config = CalendarConfig()
        self.config = config
        self.views = self._init_views()
        self._handlers = {
            CALLBACK_NEXT_MONTH: self._handle_next_month,
            CALLBACK_PREV_MONTH: self._handle_prev_month,
            CALLBACK_NEXT_YEAR: self._handle_next_year,
            CALLBACK_PREV_YEAR: self._handle_prev_year,
            CALLBACK_NEXT_YEARS_PAGE: self._handle_next_years_page,
            CALLBACK_PREV_YEARS_PAGE: self._handle_prev_years_page,
            CALLBACK_SCOPE_MONTHS: self._handle_scope_months,
            CALLBACK_SCOPE_YEARS: self._handle_scope_years,
        }

    def _init_views(self) -> Dict[CalendarScope, CalendarScopeView]:
        """
        Calendar scopes view initializer.

        Override this method customize how calendar is rendered.
        You can either set Text widgets for buttons in default views or
        create own implementation of views
        """
        return {
            CalendarScope.DAYS:
                CalendarDaysView(self._item_callback_data, self.config),
            CalendarScope.MONTHS:
                CalendarMonthView(self._item_callback_data, self.config),
            CalendarScope.YEARS:
                CalendarYearsView(self._item_callback_data, self.config),
        }

    async def _get_user_config(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> CalendarUserConfig:
        """
        User related config getter.

        Override this method to customize how user config is retrieved.

        :param data: data from window getter
        :param manager: dialog manager instance
        :return:
        """
        return CalendarUserConfig()

    async def _render_keyboard(
            self,
            data,
            manager: DialogManager,
    ) -> RawKeyboard:
        scope = self.get_scope(manager)
        view = self.views[scope]
        offset = self.get_offset(manager)
        config = await self._get_user_config(data, manager)
        if offset is None:
            offset = get_today(config.timezone)
            self.set_offset(offset, manager)
        return await view.render(config, offset, data, manager)

    def get_scope(self, manager: DialogManager) -> CalendarScope:
        calendar_data: CalendarData = self.get_widget_data(manager, {})
        current_scope = calendar_data.get("current_scope")
        if not current_scope:
            return CalendarScope.DAYS
        try:
            return CalendarScope(current_scope)
        except ValueError:
            # LOG
            return CalendarScope.DAYS

    def get_offset(self, manager: DialogManager) -> Optional[date]:
        calendar_data: CalendarData = self.get_widget_data(manager, {})
        current_offset = calendar_data.get("current_offset")
        if current_offset is None:
            return None
        return date.fromisoformat(current_offset)

    def set_offset(self, new_offset: date,
                   manager: DialogManager) -> None:
        data = self.get_widget_data(manager, {})
        data["current_offset"] = new_offset.isoformat()

    def set_scope(self, new_scope: CalendarScope,
                  manager: DialogManager) -> None:
        data = self.get_widget_data(manager, {})
        data["current_scope"] = new_scope.value

    def managed(self, manager: DialogManager) -> "ManagedCalendar":
        return ManagedCalendar(self, manager)

    async def _handle_scope_months(
            self, data: str, manager: DialogManager,
    ) -> None:
        self.set_scope(CalendarScope.MONTHS, manager)

    async def _handle_scope_years(
            self, data: str, manager: DialogManager,
    ) -> None:
        self.set_scope(CalendarScope.YEARS, manager)

    async def _handle_prev_month(
            self, data: str, manager: DialogManager,
    ) -> None:
        offset = self.get_offset(manager)
        offset = month_begin(month_begin(offset) - timedelta(days=1))
        self.set_offset(offset, manager)

    async def _handle_next_month(
            self, data: str, manager: DialogManager,
    ) -> None:
        offset = self.get_offset(manager)
        offset = next_month_begin(offset)
        self.set_offset(offset, manager)

    async def _handle_prev_year(
            self, data: str, manager: DialogManager,
    ) -> None:
        offset = self.get_offset(manager)
        offset = offset.replace(offset.year - 1)
        self.set_offset(offset, manager)

    async def _handle_next_year(
            self, data: str, manager: DialogManager,
    ) -> None:
        offset = self.get_offset(manager)
        offset = offset.replace(offset.year + 1)
        self.set_offset(offset, manager)

    async def _handle_prev_years_page(
            self, data: str, manager: DialogManager,
    ) -> None:
        offset = self.get_offset(manager)
        offset = offset.replace(offset.year - self.config.years_per_page)
        self.set_offset(offset, manager)

    async def _handle_next_years_page(
            self, data: str, manager: DialogManager,
    ) -> None:
        offset = self.get_offset(manager)
        offset = offset.replace(offset.year + self.config.years_per_page)
        self.set_offset(offset, manager)

    async def _handle_click_month(
            self, data: str, manager: DialogManager,
    ) -> None:
        offset = self.get_offset(manager)
        month = int(data[len(CALLBACK_PREFIX_MONTH):])
        offset = date(offset.year, month, 1)
        self.set_offset(offset, manager)
        self.set_scope(CalendarScope.DAYS, manager)

    async def _handle_click_year(
            self, data: str, manager: DialogManager,
    ) -> None:
        year = int(data[len(CALLBACK_PREFIX_YEAR):])
        offset = date(year, 1, 1)
        self.set_offset(offset, manager)
        self.set_scope(CalendarScope.MONTHS, manager)

    async def _handle_click_date(
            self, data: str, manager: DialogManager,
    ) -> None:
        raw_date = int(data)
        await self.on_click.process_event(
            manager.event,
            self.managed(manager),
            manager,
            date.fromtimestamp(raw_date),
        )

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        if data in self._handlers:
            handler = self._handlers[data]
        elif data.startswith(CALLBACK_PREFIX_MONTH):
            handler = self._handle_click_month
        elif data.startswith(CALLBACK_PREFIX_YEAR):
            handler = self._handle_click_year
        else:
            handler = self._handle_click_date
        await handler(data, manager)
        return True


class ManagedCalendar(ManagedWidget[Calendar]):
    def get_scope(self) -> CalendarScope:
        """Get current scope showing in calendar."""
        return self.widget.get_scope(self.manager)

    def get_offset(self) -> Optional[date]:
        """Get current offset from which calendar is shown."""
        return self.widget.get_offset(self.manager)

    def set_offset(
            self, new_offset: date,
    ) -> None:
        """Set current offset for calendar paging."""
        return self.widget.set_offset(new_offset, self.manager)

    def set_scope(
            self, new_scope: CalendarScope,
    ) -> None:
        """Set current scope showing in calendar."""
        return self.widget.set_scope(new_scope, self.manager)



from enum import Enum
from typing import Any, Optional

from aiogram.fsm.state import State
from aiogram.types import (
    Chat,
    TelegramObject,
    Update,
    User,
)
from pydantic import ConfigDict

from .modes import (
    ShowMode,
    StartMode,
)

DIALOG_EVENT_NAME = "aiogd_update"


class DialogAction(Enum):
    DONE = "DONE"
    START = "START"
    UPDATE = "UPDATE"
    SWITCH = "SWITCH"


class DialogUpdateEvent(TelegramObject):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        use_enum_values=False,
    )
    from_user: User
    chat: Chat
    action: DialogAction
    data: Any
    intent_id: Optional[str]
    stack_id: Optional[str]
    show_mode: Optional[ShowMode] = None


class DialogStartEvent(DialogUpdateEvent):
    new_state: State
    mode: StartMode


class DialogSwitchEvent(DialogUpdateEvent):
    new_state: State


class DialogUpdate(Update):
    aiogd_update: DialogUpdateEvent

    def __init__(self, aiogd_update: DialogUpdateEvent):
        super().__init__(update_id=0, aiogd_update=aiogd_update)

    @property
    def event_type(self) -> str:
        return DIALOG_EVENT_NAME

    @property
    def event(self) -> DialogUpdateEvent:
        return self.aiogd_update



from typing import Optional

from aiogram_dialog.api.entities import MediaAttachment
from aiogram_dialog.api.internal import MediaWidget
from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import BaseWidget, Whenable, WhenCondition


class Media(Whenable, BaseWidget, MediaWidget):
    def __init__(self, when: WhenCondition = None):
        super().__init__(when=when)

    async def render_media(
            self, data: dict, manager: DialogManager,
    ) -> Optional[MediaAttachment]:
        if not self.is_(data, manager):
            return None
        return await self._render_media(data, manager)

    async def _render_media(
            self, data: dict, manager: DialogManager,
    ) -> Optional[MediaAttachment]:
        return None



import dataclasses
from typing import Any, Dict, Optional

from aiogram.fsm.state import State
from aiogram.types import Message

from aiogram_dialog.api.entities import (
    ChatEvent, Data, ShowMode, StartMode,
)
from aiogram_dialog.api.entities import Context, Stack
from aiogram_dialog.api.internal import Widget
from aiogram_dialog.api.protocols import BaseDialogManager, DialogManager


class SubManager(DialogManager):
    def __init__(
            self,
            widget: Widget,
            manager: DialogManager,
            widget_id: str,
            item_id: str,
    ):
        self.widget = widget
        self.manager = manager
        self.widget_id = widget_id
        self.item_id = item_id

    @property
    def event(self) -> ChatEvent:
        return self.manager.event

    @property
    def middleware_data(self) -> Dict:
        """Middleware data."""
        return self.manager.middleware_data

    @property
    def dialog_data(self) -> Dict:
        """Dialog data for current context."""
        return self.current_context().dialog_data

    @property
    def start_data(self) -> Dict:
        """Start data for current context."""
        return self.manager.start_data

    def current_context(self) -> Context:
        context = self.manager.current_context()
        data = context.widget_data.setdefault(self.widget_id, {})
        row_data = data.setdefault(self.item_id, {})
        return dataclasses.replace(context, widget_data=row_data)

    def has_context(self) -> bool:
        return self.manager.has_context()

    def is_preview(self) -> bool:
        return self.manager.is_preview()

    def current_stack(self) -> Stack:
        return self.manager.current_stack()

    async def close_manager(self) -> None:
        return await self.manager.close_manager()

    async def show(self, show_mode: Optional[ShowMode] = None) -> Message:
        return await self.manager.show(show_mode)

    async def answer_callback(self) -> None:
        return await self.manager.answer_callback()

    async def reset_stack(self, remove_keyboard: bool = True) -> None:
        return await self.manager.reset_stack(remove_keyboard)

    async def load_data(self) -> Dict:
        return await self.manager.load_data()

    def find(self, widget_id) -> Optional[Any]:
        widget = self.widget.find(widget_id)
        if not widget:
            return None
        return widget.managed(self)

    def find_in_parent(self, widget_id) -> Optional[Any]:
        return self.manager.find(widget_id)

    @property
    def show_mode(self) -> ShowMode:
        return self.manager.show_mode

    @show_mode.setter
    def show_mode(self, show_mode: ShowMode) -> None:
        self.manager.show_mode = show_mode

    async def next(self) -> None:
        await self.manager.next()

    async def back(self) -> None:
        await self.manager.back()

    async def done(
            self,
            result: Any = None,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        await self.manager.done(result, show_mode)

    async def mark_closed(self) -> None:
        await self.manager.mark_closed()

    async def start(self, state: State, data: Data = None,
                    mode: StartMode = StartMode.NORMAL,
                    show_mode: Optional[ShowMode] = None) -> None:
        await self.manager.start(
            state=state, data=data, mode=mode, show_mode=show_mode,
        )

    async def switch_to(
            self,
            state: State,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        await self.manager.switch_to(state, show_mode)

    async def update(
            self,
            data: Dict,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        self.current_context().dialog_data.update(data)
        await self.show(show_mode)

    def bg(self, user_id: Optional[int] = None, chat_id: Optional[int] = None,
           stack_id: Optional[str] = None,
           load: bool = False) -> BaseDialogManager:
        return self.manager.bg(
            user_id=user_id, chat_id=chat_id, stack_id=stack_id, load=load,
        )



from abc import abstractmethod
from typing import Dict, Optional, Protocol, Union

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.internal import RawKeyboard
from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import ManagedWidget, WhenCondition
from aiogram_dialog.widgets.kbd.base import Keyboard
from aiogram_dialog.widgets.text import Const, Format, Text
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor,
    WidgetEventProcessor,
)


class OnCounterEvent(Protocol):
    @abstractmethod
    async def __call__(
            self,
            event: ChatEvent,
            counter: "ManagedCounter",  # noqa: F841
            dialog_manager: DialogManager,
    ):
        raise NotImplementedError


OnCounterEventVariant = Union[OnCounterEvent, WidgetEventProcessor, None]

PLUS_TEXT = Const("+")
MINUS_TEXT = Const("-")
DEFAULT_COUNTER_TEXT = Format("{value:g}")


class Counter(Keyboard):
    """
    Counter widget.

    Used to represent number with increment/decrement buttons
    To remove any button set its text to `None`
    """

    def __init__(
            self,
            id: str,
            plus: Optional[Text] = PLUS_TEXT,
            minus: Optional[Text] = MINUS_TEXT,
            text: Optional[Text] = DEFAULT_COUNTER_TEXT,
            min_value: float = 0,
            max_value: float = 999999,
            increment: float = 1,
            default: float = 0,
            cycle: bool = False,
            on_click: OnCounterEventVariant = None,
            on_text_click: OnCounterEventVariant = None,
            on_value_changed: OnCounterEventVariant = None,
            when: WhenCondition = None,
    ) -> None:
        """
        Init counter widget.

        :param id: ID of widget
        :param plus: Text to render `+`-button. Set `None` to disable
        :param minus: Text to render `-`-button. Set `None` to disable
        :param text: Text to render button with current value. \
        Set `None` to disable
        :param min_value: Minimal allowed value
        :param max_value: Maximum allowed value
        :param increment: Step used to increment
        :param default: Default value
        :param cycle: Whether cycle values on overflow
        :param on_click: Callback to process any click
        :param on_text_click: Callback to process click on `text`-button
        :param on_value_changed: Callback to process value changes, \
        regardless of the reason
        :param when: Condition when to show widget
        """
        super().__init__(id=id, when=when)
        self.plus = plus
        self.minus = minus
        self.min = min_value
        self.max = max_value
        self.increment = increment
        self.default = default
        self.text = text
        self.cycle = cycle
        self.on_click = ensure_event_processor(on_click)
        self.on_value_changed = ensure_event_processor(on_value_changed)
        self.on_text_click = ensure_event_processor(on_text_click)

    def get_value(self, manager: DialogManager) -> float:
        return self.get_widget_data(manager, self.default)

    async def set_value(self, manager: DialogManager,
                        value: float) -> None:
        if self.min <= value <= self.max:
            self.set_widget_data(manager, value)
            await self.on_value_changed.process_event(
                manager.event,
                self.managed(manager),
                manager,
            )

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> RawKeyboard:
        row = []
        if self.minus:
            minus = await self.minus.render_text(data, manager)
            row.append(
                InlineKeyboardButton(
                    text=minus,
                    callback_data=self._item_callback_data("-"),
                ),
            )
        if self.text:
            text = await self.text.render_text(
                {"value": self.get_value(manager), "data": data},
                manager,
            )
            row.append(
                InlineKeyboardButton(
                    text=text,
                    callback_data=self._item_callback_data(""),
                ),
            )
        if self.plus:
            plus = await self.plus.render_text(data, manager)
            row.append(
                InlineKeyboardButton(
                    text=plus,
                    callback_data=self._item_callback_data("+"),
                ),
            )
        return [row]

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        await self.on_click.process_event(
            callback, self.managed(manager), manager,
        )

        value = self.get_value(manager)
        if data == "+":
            value += self.increment
            if value > self.max and self.cycle:
                value = self.min
            await self.set_value(manager, value)
        elif data == "-":
            value -= self.increment
            if value < self.min and self.cycle:
                value = self.max
            await self.set_value(manager, value)
        elif data == "":
            await self.on_text_click.process_event(
                callback, self.managed(manager), manager,
            )
        return True

    def managed(self, manager: DialogManager):
        return ManagedCounter(self, manager)


class ManagedCounter(ManagedWidget[Counter]):
    def get_value(self) -> float:
        """Get current value set in counter."""
        return self.widget.get_value(self.manager)

    async def set_value(self, value: float) -> None:
        """Change current counter value."""
        await self.widget.set_value(self.manager, value)



from abc import abstractmethod
from typing import (
    Any, Awaitable, Callable, Dict, List, Optional, Protocol,
    runtime_checkable, Union,
)

from aiogram.types import (
    CallbackQuery, InlineKeyboardButton, KeyboardButton, Message,
)

from aiogram_dialog import DialogManager
from aiogram_dialog.api.entities import MarkupVariant, MediaAttachment
from aiogram_dialog.api.protocols import DialogProtocol


@runtime_checkable
class Widget(Protocol):
    @abstractmethod
    def managed(self, manager: DialogManager) -> Any:
        raise NotImplementedError

    @abstractmethod
    def find(self, widget_id: str) -> Optional["Widget"]:
        raise NotImplementedError


@runtime_checkable
class TextWidget(Widget, Protocol):
    @abstractmethod
    async def render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        """Create text."""
        raise NotImplementedError


ButtonVariant = Union[InlineKeyboardButton, KeyboardButton]
RawKeyboard = List[List[ButtonVariant]]


@runtime_checkable
class KeyboardWidget(Widget, Protocol):
    @abstractmethod
    async def render_keyboard(
            self, data: Dict, manager: DialogManager,
    ) -> RawKeyboard:
        """Create Inline keyboard contents."""
        raise NotImplementedError

    @abstractmethod
    async def process_callback(
            self, callback: CallbackQuery, dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        """
        Handle user click on some inline button.

        Invoked regardless if callback belongs to current widget.

        returns True if callback processed and should not be propagated
        """
        raise NotImplementedError


@runtime_checkable
class MediaWidget(Widget, Protocol):
    @abstractmethod
    async def render_media(
            self, data: dict, manager: DialogManager,
    ) -> Optional[MediaAttachment]:
        """Create media attachment."""
        raise NotImplementedError


@runtime_checkable
class InputWidget(Widget, Protocol):
    @abstractmethod
    async def process_message(
            self, message: Message, dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        """
        Handle incoming message from user.

        Invoked regardless if callback belongs to current widget.

        returns True if callback processed and should not be propagated
        """
        raise NotImplementedError


DataGetter = Callable[..., Awaitable[Dict]]


@runtime_checkable
class MarkupFactory(Protocol):
    @abstractmethod
    async def render_markup(
            self, data: dict, manager: DialogManager, keyboard: RawKeyboard,
    ) -> MarkupVariant:
        """Render reply_markup using prepared keyboard."""
        raise NotImplementedError



from copy import copy
from typing import Dict, Optional, Type

from aiogram import Bot
from aiogram.fsm.state import State, StatesGroup
from aiogram.fsm.storage.base import BaseStorage, StorageKey

from aiogram_dialog.api.entities import (
    Context, DEFAULT_STACK_ID, Stack,
)
from aiogram_dialog.api.exceptions import UnknownIntent, UnknownState


class StorageProxy:
    def __init__(
            self,
            storage: BaseStorage,
            user_id: int,
            chat_id: int,
            bot: Bot,
            state_groups: Dict[str, Type[StatesGroup]],
    ):
        self.storage = storage
        self.state_groups = state_groups
        self.user_id = user_id
        self.chat_id = chat_id
        self.bot = bot

    async def load_context(self, intent_id: str) -> Context:
        data = await self.storage.get_data(
            key=self._context_key(intent_id),
        )
        if not data:
            raise UnknownIntent(
                f"Context not found for intent id: {intent_id}",
            )
        data["state"] = self._state(data["state"])
        return Context(**data)

    async def load_stack(self, stack_id: str = DEFAULT_STACK_ID) -> Stack:
        data = await self.storage.get_data(
            key=self._stack_key(stack_id),
        )
        if not data:
            return Stack(_id=stack_id)
        return Stack(**data)

    async def save_context(self, context: Optional[Context]) -> None:
        if not context:
            return
        data = copy(vars(context))
        data["state"] = data["state"].state
        await self.storage.set_data(
            key=self._context_key(context.id),
            data=data,
        )

    async def remove_context(self, intent_id: str):
        await self.storage.set_data(
            key=self._context_key(intent_id),
            data={},
        )

    async def remove_stack(self, stack_id: str):
        await self.storage.set_data(
            key=self._stack_key(stack_id),
            data={},
        )

    async def save_stack(self, stack: Optional[Stack]) -> None:
        if not stack:
            return
        if stack.empty() and not stack.last_message_id:
            await self.storage.set_data(
                key=self._stack_key(stack.id),
                data={},
            )
        else:
            data = copy(vars(stack))
            await self.storage.set_data(
                key=self._stack_key(stack.id),
                data=data,
            )

    def _context_key(self, intent_id: str) -> StorageKey:
        return StorageKey(
            bot_id=self.bot.id,
            chat_id=self.chat_id,
            user_id=self.user_id,
            destiny=f"aiogd:context:{intent_id}",
        )

    def _stack_key(self, stack_id: str) -> StorageKey:
        return StorageKey(
            bot_id=self.bot.id,
            chat_id=self.chat_id,
            user_id=self.user_id,
            destiny=f"aiogd:stack:{stack_id}",
        )

    def _state(self, state: str) -> State:
        group, *_ = state.partition(":")
        try:
            for real_state in self.state_groups[group].__all_states__:
                if real_state.state == state:
                    return real_state
        except KeyError:
            raise UnknownState(f"Unknown state group {group}")
        raise UnknownState(f"Unknown state {state}")



from abc import abstractmethod
from typing import Optional, Union

from aiogram.types import CallbackQuery

from aiogram_dialog.api.internal import KeyboardWidget, RawKeyboard
from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import (
    Actionable, Whenable, WhenCondition,
)


class Keyboard(Actionable, Whenable, KeyboardWidget):
    def __init__(self, id: Optional[str] = None, when: WhenCondition = None):
        Actionable.__init__(self, id=id)
        Whenable.__init__(self, when=when)

    async def render_keyboard(
            self,
            data,
            manager: DialogManager,
    ) -> RawKeyboard:
        """
        Create inline keyboard contents.

        When inheriting override `_render_keyboard` method instead
        if you want to keep processing of `when` condition
        """
        if not self.is_(data, manager):
            return []
        return await self._render_keyboard(data, manager)

    @abstractmethod
    async def _render_keyboard(
            self,
            data: dict,
            manager: DialogManager,
    ) -> RawKeyboard:
        """
        Create inline keyboard contents.

        Called if widget is not hidden only (regarding `when`-condition)
        """
        raise NotImplementedError

    def callback_prefix(self):
        if not self.widget_id:
            return None
        return f"{self.widget_id}:"

    def _own_callback_data(self) -> Union[str, None]:
        """Create callback data for only button in widget."""
        return self.widget_id

    def _item_callback_data(self, data: Union[str, int]):
        """Create callback data for widgets button if multiple."""
        return f"{self.callback_prefix()}{data}"

    async def process_callback(
            self,
            callback: CallbackQuery,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        if callback.data == self.widget_id:
            return await self._process_own_callback(
                callback,
                dialog,
                manager,
            )
        prefix = self.callback_prefix()
        if prefix and callback.data.startswith(prefix):
            return await self._process_item_callback(
                callback,
                callback.data[len(prefix):],
                dialog,
                manager,
            )
        return await self._process_other_callback(callback, dialog, manager)

    async def _process_own_callback(
            self,
            callback: CallbackQuery,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        """Process callback related to _own_callback_data."""
        return False

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        """Process callback related to _item_callback_data."""
        return False

    async def _process_other_callback(
            self,
            callback: CallbackQuery,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        """
        Process callback for unknown callback data.

        Can be used for layouts
        """
        return False



from typing import Callable, Dict, Union

from aiogram.types import KeyboardButton

from aiogram_dialog.api.internal import RawKeyboard
from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.text import Text
from .base import Keyboard


class RequestContact(Keyboard):
    def __init__(
            self,
            text: Text,
            when: Union[str, Callable, None] = None,
    ):
        super().__init__(when=when)
        self.text = text

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> RawKeyboard:
        return [
            [
                KeyboardButton(
                    text=await self.text.render_text(data, manager),
                    request_contact=True,
                ),
            ],
        ]


class RequestLocation(Keyboard):
    def __init__(
            self,
            text: Text,
            when: Union[str, Callable, None] = None,
    ):
        super().__init__(when=when)
        self.text = text

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> RawKeyboard:
        return [
            [
                KeyboardButton(
                    text=await self.text.render_text(data, manager),
                    request_location=True,
                ),
            ],
        ]



from datetime import datetime
from logging import getLogger
from typing import Any, Dict, Optional

from aiogram import Router
from aiogram.fsm.state import State
from aiogram.types import (
    CallbackQuery, Chat, Message, ReplyKeyboardMarkup, User,
)

from aiogram_dialog.api.entities import (
    ChatEvent, Context, Data, DEFAULT_STACK_ID, LaunchMode, MediaId,
    NewMessage, ShowMode, Stack, StartMode,
)
from aiogram_dialog.api.entities import OldMessage, UnknownText
from aiogram_dialog.api.exceptions import (
    IncorrectBackgroundError, InvalidKeyboardType, NoContextError,
)
from aiogram_dialog.api.internal import (
    CONTEXT_KEY, STACK_KEY, STORAGE_KEY,
)
from aiogram_dialog.api.internal import (
    FakeChat, FakeUser,
)
from aiogram_dialog.api.protocols import (
    BaseDialogManager, DialogManager, DialogProtocol, DialogRegistryProtocol,
    MediaIdStorageProtocol, MessageManagerProtocol, MessageNotModified,
)
from aiogram_dialog.context.storage import StorageProxy
from aiogram_dialog.utils import get_media_id
from .bg_manager import BgManager

logger = getLogger(__name__)


class ManagerImpl(DialogManager):

    def __init__(
            self, event: ChatEvent,
            message_manager: MessageManagerProtocol,
            media_id_storage: MediaIdStorageProtocol,
            registry: DialogRegistryProtocol,
            router: Router,
            data: Dict,
    ):
        self.disabled = False
        self.message_manager = message_manager
        self.media_id_storage = media_id_storage
        self._event = event
        self._data = data
        self._show_mode: ShowMode = ShowMode.AUTO
        self._registry = registry
        self._router = router

    @property
    def show_mode(self) -> ShowMode:
        """Get current show mode, used for next show action."""
        return self._show_mode

    @show_mode.setter
    def show_mode(self, show_mode: ShowMode) -> None:
        """Set current show mode, used for next show action."""
        self._show_mode = show_mode

    @property
    def event(self) -> ChatEvent:
        return self._event

    @property
    def middleware_data(self) -> Dict:
        """Middleware data."""
        return self._data

    @property
    def dialog_data(self) -> Dict:
        """Dialog data for current context."""
        return self.current_context().dialog_data

    @property
    def start_data(self) -> Dict:
        """Start data for current context."""
        return self.current_context().start_data

    def check_disabled(self):
        if self.disabled:
            raise IncorrectBackgroundError(
                "Detected background access to dialog manager. "
                "Please use background manager available via `manager.bg()` "
                "method to access methods from background tasks",
            )

    async def load_data(self) -> Dict:
        context = self.current_context()
        return {
            "dialog_data": context.dialog_data,
            "start_data": context.start_data,
            "middleware_data": self._data,
            "event": self.event,
        }

    def is_preview(self) -> bool:
        return False

    def dialog(self) -> DialogProtocol:
        self.check_disabled()
        current = self.current_context()
        if not current:
            raise RuntimeError
        return self._registry.find_dialog(current.state)

    def current_context(self) -> Context:
        self.check_disabled()
        context = self._current_context_unsafe()
        if not context:
            logger.warning(
                "Trying to access current context, while no dialog is opened",
            )
            raise NoContextError
        return context

    def _current_context_unsafe(self) -> Optional[Context]:
        return self._data[CONTEXT_KEY]

    def has_context(self) -> bool:
        self.check_disabled()
        return bool(self._current_context_unsafe())

    def current_stack(self) -> Stack:
        self.check_disabled()
        return self._data[STACK_KEY]

    def storage(self) -> StorageProxy:
        return self._data[STORAGE_KEY]

    async def _remove_kbd(self) -> None:
        if self.current_stack().last_message_id is None:
            return
        chat = self._data["event_chat"]
        bot = self._data["bot"]
        message = Message(
            chat=chat,
            message_id=self.current_stack().last_message_id,
            date=datetime.now(),
        )
        await self.message_manager.remove_kbd(
            bot,
            message,
        )
        self.current_stack().last_message_id = None

    async def _process_last_dialog_result(
            self,
            start_data: Data,
            result: Any,
    ) -> None:
        """Process closing last dialog in stack."""
        await self._remove_kbd()

    async def done(
            self,
            result: Any = None,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        self.check_disabled()
        await self.dialog().process_close(result, self)
        old_context = self.current_context()
        await self.mark_closed()
        context = self._current_context_unsafe()
        if not context:
            await self._process_last_dialog_result(
                old_context.start_data,
                result,
            )
            return
        dialog = self.dialog()
        await dialog.process_result(old_context.start_data, result, self)
        new_context = self._current_context_unsafe()
        if new_context and context.id == new_context.id:
            await self.show(show_mode)

    async def answer_callback(self) -> None:
        if not isinstance(self.event, CallbackQuery):
            return
        return await self.message_manager.answer_callback(
            bot=self._data["bot"],
            callback_query=self.event,
        )

    async def mark_closed(self) -> None:
        self.check_disabled()
        storage = self.storage()
        stack = self.current_stack()
        await storage.remove_context(stack.pop())
        if stack.empty():
            self._data[CONTEXT_KEY] = None
        else:
            intent_id = stack.last_intent_id()
            self._data[CONTEXT_KEY] = await storage.load_context(intent_id)
        await storage.save_stack(stack)

    async def start(
            self,
            state: State,
            data: Data = None,
            mode: StartMode = StartMode.NORMAL,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        self.check_disabled()
        self.show_mode = show_mode or self.show_mode
        if mode is StartMode.NORMAL:
            await self._start_normal(state, data)
        elif mode is StartMode.RESET_STACK:
            await self.reset_stack(remove_keyboard=False)
            await self._start_normal(state, data)
        elif mode is StartMode.NEW_STACK:
            await self._start_new_stack(state, data)
        else:
            raise ValueError(f"Unknown start mode: {mode}")

    async def reset_stack(self, remove_keyboard: bool = True) -> None:
        self.check_disabled()
        storage = self.storage()
        stack = self.current_stack()
        while not stack.empty():
            await storage.remove_context(stack.pop())
        await storage.save_stack(stack)
        if remove_keyboard:
            await self._remove_kbd()
        self._data[CONTEXT_KEY] = None

    async def _start_new_stack(self, state: State, data: Data = None) -> None:
        stack = Stack()
        await self.bg(stack_id=stack.id).start(
            state, data, StartMode.NORMAL, self.show_mode,
        )

    async def _start_normal(self, state: State, data: Data = None) -> None:
        stack = self.current_stack()
        old_dialog: Optional[DialogProtocol] = None
        if not stack.empty():
            old_dialog = self.dialog()
            if old_dialog.launch_mode is LaunchMode.EXCLUSIVE:
                raise ValueError(
                    "Cannot start dialog on top "
                    "of one with launch_mode==SINGLE",
                )

        new_dialog = self._registry.find_dialog(state)
        launch_mode = new_dialog.launch_mode
        if launch_mode in (LaunchMode.EXCLUSIVE, LaunchMode.ROOT):
            await self.reset_stack(remove_keyboard=False)
        if launch_mode is LaunchMode.SINGLE_TOP:
            if new_dialog is old_dialog:
                await self.storage().remove_context(stack.pop())

        if self.has_context():
            await self.storage().save_context(self.current_context())
        context = stack.push(state, data)
        self._data[CONTEXT_KEY] = context
        await self.dialog().process_start(self, data, state)
        new_context = self._current_context_unsafe()
        if new_context and context.id == new_context.id:
            await self.show()

    async def next(self) -> None:
        context = self.current_context()
        states = self.dialog().states()
        current_index = states.index(context.state)
        new_state = states[current_index + 1]
        await self.switch_to(new_state)

    async def back(self) -> None:
        context = self.current_context()
        states = self.dialog().states()
        current_index = states.index(context.state)
        new_state = states[current_index - 1]
        await self.switch_to(new_state)

    async def switch_to(
            self,
            state: State,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        self.check_disabled()
        context = self.current_context()
        if context.state.group != state.group:
            raise ValueError(
                f"Cannot switch to another state group. "
                f"Current state: {context.state}, asked for {state}",
            )
        self.show_mode = show_mode or self.show_mode
        context.state = state

    def _ensure_stack_compatible(
            self, stack: Stack, new_message: NewMessage,
    ) -> None:
        if stack.id == DEFAULT_STACK_ID:
            return  # no limitations for default stack
        if isinstance(new_message.reply_markup, ReplyKeyboardMarkup):
            raise InvalidKeyboardType(
                "Cannot use ReplyKeyboardMarkup in non default stack",
            )

    async def show(self, show_mode: Optional[ShowMode] = None) -> None:
        stack = self.current_stack()
        bot = self._data["bot"]
        old_message = self._get_last_message()
        if self.show_mode is ShowMode.NO_UPDATE:
            logger.debug("ShowMode is NO_UPDATE, skip rendering")
            return

        new_message = await self.dialog().render(self)
        new_message.show_mode = show_mode or self.show_mode
        if new_message.show_mode is ShowMode.AUTO:
            new_message.show_mode = self._calc_show_mode()
        await self._fix_cached_media_id(new_message)

        self._ensure_stack_compatible(stack, new_message)

        try:
            sent_message = await self.message_manager.show_message(
                bot, new_message, old_message,
            )
        except MessageNotModified:
            # nothing changed so nothing to save
            # we do not have the actual version of message
            logger.debug("MessageNotModified, not storing ids")
        else:
            self._save_last_message(sent_message)
            if new_message.media:
                await self.media_id_storage.save_media_id(
                    path=new_message.media.path,
                    url=new_message.media.url,
                    type=new_message.media.type,
                    media_id=MediaId(
                        sent_message.media_id,
                        sent_message.media_uniq_id,
                    ),
                )
        if isinstance(self.event, Message):
            stack.last_income_media_group_id = self.event.media_group_id

    async def _fix_cached_media_id(self, new_message: NewMessage):
        if not new_message.media or new_message.media.file_id:
            return
        new_message.media.file_id = await self.media_id_storage.get_media_id(
            path=new_message.media.path,
            url=new_message.media.url,
            type=new_message.media.type,
        )

    def _get_last_message(self) -> Optional[OldMessage]:
        stack = self.current_stack()
        chat = self._data["event_chat"]
        if (
                isinstance(self.event, CallbackQuery) and
                self.event.message and
                stack.last_message_id == self.event.message.message_id
        ):
            current_message = self.event.message
            media_id = get_media_id(current_message)
            return OldMessage(
                media_id=(media_id.file_id if media_id else None),
                media_uniq_id=(media_id.file_unique_id if media_id else None),
                text=current_message.text,
                has_reply_keyboard=False,
                chat=chat,
                message_id=current_message.message_id,
            )
        if not stack or not stack.last_message_id:
            return None
        return OldMessage(
            media_id=stack.last_media_id,
            media_uniq_id=stack.last_media_unique_id,
            text=UnknownText.UNKNOWN,
            has_reply_keyboard=stack.last_reply_keyboard,
            chat=chat,
            message_id=stack.last_message_id,
        )

    def _save_last_message(self, message: OldMessage):
        stack = self.current_stack()
        stack.last_message_id = message.message_id
        stack.last_media_id = message.media_id
        stack.last_media_unique_id = message.media_uniq_id
        stack.last_reply_keyboard = message.has_reply_keyboard

    def _calc_show_mode(self) -> ShowMode:
        if self.show_mode is not ShowMode.AUTO:
            return self.show_mode
        if self.current_stack().last_reply_keyboard:
            return ShowMode.DELETE_AND_SEND
        if self.current_stack().id != DEFAULT_STACK_ID:
            return ShowMode.EDIT
        if isinstance(self.event, Message):
            if self.event.media_group_id is None:
                return ShowMode.SEND
            elif self.event.media_group_id == \
                    self.current_stack().last_income_media_group_id:
                return ShowMode.EDIT
            else:
                return ShowMode.SEND
        return ShowMode.EDIT

    async def update(
            self,
            data: Dict,
            show_mode: Optional[ShowMode] = None,
    ) -> None:
        self.current_context().dialog_data.update(data)
        await self.show(show_mode)

    def find(self, widget_id) -> Optional[Any]:
        widget = self.dialog().find(widget_id)
        if not widget:
            return None
        return widget.managed(self)

    def is_same_chat(self, user: User, chat: Chat) -> bool:
        if "event_chat" not in self._data:
            return False

        current_chat = self._data["event_chat"]
        current_user = self.event.from_user
        return user.id == current_user.id and chat.id == current_chat.id

    def _get_fake_user(self, user_id: Optional[int] = None) -> User:
        """Get User if we have info about him or FakeUser instead."""
        current_user = self.event.from_user
        if user_id in (None, current_user.id):
            return current_user
        return FakeUser(id=user_id, is_bot=False, first_name="")

    def _get_fake_chat(self, chat_id: Optional[int] = None) -> Chat:
        """Get Chat if we have info about him or FakeChat instead."""
        if "event_chat" in self._data:
            current_chat = self._data["event_chat"]
            if chat_id in (None, current_chat.id):
                return current_chat
        elif chat_id is None:
            raise ValueError(
                "Explicit `chat_id` is required "
                "for events without current chat",
            )
        return FakeChat(id=chat_id, type="")

    def bg(
            self,
            user_id: Optional[int] = None,
            chat_id: Optional[int] = None,
            stack_id: Optional[str] = None,
            load: bool = False,
    ) -> BaseDialogManager:
        user = self._get_fake_user(user_id)
        chat = self._get_fake_chat(chat_id)
        intent_id = None
        if stack_id is None:
            if self.is_same_chat(user, chat):
                stack_id = self.current_stack().id
                if self.has_context():
                    intent_id = self.current_context().id
            else:
                stack_id = DEFAULT_STACK_ID

        return BgManager(
            user=user,
            chat=chat,
            bot=self._data["bot"],
            router=self._router,
            intent_id=intent_id,
            stack_id=stack_id,
            load=load,
        )

    async def close_manager(self) -> None:
        self.check_disabled()
        self.disabled = True
        del self.media_id_storage
        del self.message_manager
        del self._event
        del self._data



__all__ = [
    "Keyboard",
    "Button",
    "Url",
    "WebApp",
    "SwitchInlineQuery",
    "Calendar",
    "CalendarConfig",
    "CalendarScope",
    "CalendarUserConfig",
    "ManagedCalendar",
    "Counter",
    "ManagedCounter",
    "Back",
    "Cancel",
    "Next",
    "Start",
    "SwitchTo",
    "Group",
    "Row",
    "Column",
    "CurrentPage",
    "FirstPage",
    "LastPage",
    "NextPage",
    "NumberedPager",
    "PrevPage",
    "SwitchPage",
    "ScrollingGroup",
    "RequestContact",
    "RequestLocation",
    "Checkbox",
    "ManagedCheckbox",
    "Select",
    "Radio",
    "Multiselect",
    "ManagedMultiselect",
    "ManagedRadio",
    "ListGroup",
    "ManagedListGroup",
    "StubScroll",
]

from .base import Keyboard
from .button import Button, SwitchInlineQuery, Url, WebApp
from .calendar_kbd import (
    Calendar, CalendarConfig, CalendarScope, CalendarUserConfig,
    ManagedCalendar,
)
from .checkbox import Checkbox, ManagedCheckbox
from .counter import Counter, ManagedCounter
from .group import Column, Group, Row
from .list_group import ListGroup, ManagedListGroup
from .pager import (
    CurrentPage,
    FirstPage,
    LastPage,
    NextPage,
    NumberedPager,
    PrevPage,
    SwitchPage,
)
from .request import RequestContact, RequestLocation
from .scrolling_group import ScrollingGroup
from .select import (
    ManagedMultiselect,
    ManagedRadio,
    Multiselect,
    Radio,
    Select,
)
from .state import Back, Cancel, Next, Start, SwitchTo
from .stub_scroll import StubScroll



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from aiogram.types import (
    Chat, ForceReply, InlineKeyboardMarkup, ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
)

from aiogram_dialog.api.entities import MediaAttachment, ShowMode

MarkupVariant = Union[
    ForceReply, InlineKeyboardMarkup, ReplyKeyboardMarkup, ReplyKeyboardRemove,
]


class UnknownText(Enum):
    UNKNOWN = object()


@dataclass
class OldMessage:
    chat: Chat
    message_id: int
    media_id: Optional[str]
    media_uniq_id: Optional[str]
    text: Union[str, None, UnknownText] = None
    has_reply_keyboard: bool = False


@dataclass
class NewMessage:
    chat: Chat
    text: Optional[str] = None
    reply_markup: Optional[MarkupVariant] = None
    parse_mode: Optional[str] = None
    show_mode: ShowMode = ShowMode.AUTO
    disable_web_page_preview: Optional[bool] = None
    media: Optional[MediaAttachment] = None



from logging import getLogger
from typing import List, Optional, Tuple, Union

from aiogram.types import (
    CallbackQuery, Chat, ChatMemberUpdated, InlineKeyboardButton,
    KeyboardButton, Message, User,
)

from aiogram_dialog.api.entities import (
    ChatEvent, DialogUpdateEvent, MediaId, NewMessage,
)

logger = getLogger(__name__)

CB_SEP = "\x1D"

REPLY_CALLBACK_SYMBOLS: str = (
    "\u200C"
    "\u200D"
    "\u2060"
    "\u2061"
    "\u2062"
    "\u2063"
    "\u2064"
    "\u00AD"
    "\U0001D173"
    "\U0001D174"
    "\U0001D175"
    "\U0001D176"
    "\U0001D177"
    "\U0001D178"
    "\U0001D179"
    "\U0001D17A"
)


def _encode_reply_callback_byte(byte: int):
    return (
        REPLY_CALLBACK_SYMBOLS[byte % len(REPLY_CALLBACK_SYMBOLS)] +
        REPLY_CALLBACK_SYMBOLS[byte // len(REPLY_CALLBACK_SYMBOLS)]
    )


def encode_reply_callback(data: str) -> str:
    bytes_data = data.encode("utf-8")
    return "".join(
        _encode_reply_callback_byte(byte)
        for byte in bytes_data
    )


def _decode_reply_callback_byte(little: str, big: str) -> int:
    return (
        REPLY_CALLBACK_SYMBOLS.index(big) * len(REPLY_CALLBACK_SYMBOLS) +
        REPLY_CALLBACK_SYMBOLS.index(little)
    )


def join_reply_callback(text: str, callback_data: str) -> str:
    return text + encode_reply_callback(callback_data)


def split_reply_callback(
        data: Optional[str],
) -> Tuple[Optional[str], Optional[str]]:
    if not data:
        return None, None
    text = data.rstrip(REPLY_CALLBACK_SYMBOLS)
    callback = data[len(text):]
    return text, decode_reply_callback(callback)


def decode_reply_callback(data: str) -> str:
    bytes_data = bytes(
        _decode_reply_callback_byte(little, big)
        for little, big in zip(data[::2], data[1::2])
    )
    return bytes_data.decode("utf-8")


def _transform_to_reply_button(
        button: Union[InlineKeyboardButton, KeyboardButton],
) -> KeyboardButton:
    if isinstance(button, KeyboardButton):
        return button
    if not button.callback_data:
        raise ValueError("Cannot convert inline button without callback_data")
    return KeyboardButton(text=join_reply_callback(
        text=button.text, callback_data=button.callback_data,
    ))


def transform_to_reply_keyboard(
        keyboard: List[List[Union[InlineKeyboardButton, KeyboardButton]]],
) -> List[List[KeyboardButton]]:
    new_kdb = []
    for row in keyboard:
        new_row = []
        new_kdb.append(new_row)
        for button in row:
            new_row.append(_transform_to_reply_button(button))
    return new_kdb


def get_chat(event: ChatEvent) -> Chat:
    if isinstance(event, (Message, DialogUpdateEvent, ChatMemberUpdated)):
        return event.chat
    elif isinstance(event, CallbackQuery):
        if not event.message:
            return Chat(id=event.from_user.id, type="")
        return event.message.chat


def is_chat_loaded(chat: Chat) -> bool:
    """
    Check if chat is correctly loaded from telegram.

    For internal events it can be created with no data inside as a FakeChat
    """
    return not getattr(chat, "fake", False)


def is_user_loaded(user: User) -> bool:
    """
    Check if user is correctly loaded from telegram.

    For internal events it can be created with no data inside as a FakeUser
    """
    return not getattr(user, "fake", False)


def get_media_id(message: Message) -> Optional[MediaId]:
    media = (
        message.audio or
        message.animation or
        message.document or
        (message.photo[-1] if message.photo else None) or
        message.video
    )
    if not media:
        return None
    return MediaId(
        file_id=media.file_id,
        file_unique_id=media.file_unique_id,
    )


def intent_callback_data(
        intent_id: str, callback_data: Optional[str],
) -> Optional[str]:
    if callback_data is None:
        return None
    prefix = intent_id + CB_SEP
    if callback_data.startswith(prefix):
        return callback_data
    return prefix + callback_data


def add_indent_id(message: NewMessage, intent_id: str):
    if not message.reply_markup:
        return
    inline_keyboard = getattr(message.reply_markup, "inline_keyboard", None)
    if not inline_keyboard:
        return
    for row in inline_keyboard:
        for button in row:
            button.callback_data = intent_callback_data(
                intent_id, button.callback_data,
            )


def remove_indent_id(callback_data: str) -> Tuple[Optional[str], str]:
    if CB_SEP in callback_data:
        intent_id, new_data = callback_data.split(CB_SEP, maxsplit=1)
        return intent_id, new_data
    return None, callback_data



__all__ = [
    "Actionable",
    "BaseWidget",
    "ManagedWidget",
    "BaseScroll", "ManagedScroll",
    "OnPageChanged", "OnPageChangedVariants", "Scroll",
    "true_condition", "Whenable", "WhenCondition",
]

from .action import Actionable
from .base import BaseWidget
from .managed import ManagedWidget
from .scroll import (
    BaseScroll, ManagedScroll, OnPageChanged, OnPageChangedVariants, Scroll,
)
from .when import true_condition, Whenable, WhenCondition



from abc import abstractmethod
from typing import Dict, Protocol, Type, Union

from aiogram.fsm.state import State, StatesGroup

from .dialog import DialogProtocol


class DialogRegistryProtocol(Protocol):
    @abstractmethod
    def find_dialog(self, state: Union[State, str]) -> DialogProtocol:
        raise NotImplementedError

    @abstractmethod
    def state_groups(self) -> Dict[str, Type[StatesGroup]]:
        raise NotImplementedError



from typing import Union

from aiogram.types import CallbackQuery, ChatMemberUpdated, Message

from .update_event import DialogUpdateEvent

ChatEvent = Union[CallbackQuery, Message, DialogUpdateEvent, ChatMemberUpdated]



from logging import getLogger
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from aiogram import Router
from aiogram.fsm.state import State, StatesGroup
from aiogram.types import CallbackQuery, Message

from aiogram_dialog.api.entities import Data, LaunchMode, NewMessage
from aiogram_dialog.api.exceptions import (
    UnregisteredWindowError,
)
from aiogram_dialog.api.internal import Widget, WindowProtocol
from aiogram_dialog.api.protocols import (
    DialogManager, DialogProtocol,
)
from .context.intent_filter import IntentFilter
from .utils import add_indent_id, remove_indent_id, split_reply_callback
from .widgets.data import PreviewAwareGetter
from .widgets.utils import ensure_data_getter, GetterVariant

logger = getLogger(__name__)

ChatEvent = Union[CallbackQuery, Message]
OnDialogEvent = Callable[[Any, DialogManager], Awaitable]
OnResultEvent = Callable[[Data, Any, DialogManager], Awaitable]
W = TypeVar("W", bound=Widget)


class Dialog(Router, DialogProtocol):
    def __init__(
            self,
            *windows: WindowProtocol,
            on_start: Optional[OnDialogEvent] = None,
            on_close: Optional[OnDialogEvent] = None,
            on_process_result: Optional[OnResultEvent] = None,
            launch_mode: LaunchMode = LaunchMode.STANDARD,
            getter: GetterVariant = None,
            preview_data: GetterVariant = None,
            name: Optional[str] = None,
    ):
        super().__init__(name=name or windows[0].get_state().group.__name__)
        self._states_group = windows[0].get_state().group
        self._states: List[State] = []
        for w in windows:
            if w.get_state().group != self._states_group:
                raise ValueError(
                    "All windows must be attached to same StatesGroup",
                )
            state = w.get_state()
            if state in self._states:
                raise ValueError(f"Multiple windows with state {state}")
            self._states.append(state)
        self.windows: Dict[State, WindowProtocol] = dict(
            zip(self._states, windows),
        )
        self.on_start = on_start
        self.on_close = on_close
        self.on_process_result = on_process_result
        self._launch_mode = launch_mode
        self.getter = PreviewAwareGetter(
            ensure_data_getter(getter),
            ensure_data_getter(preview_data),
        )
        self._setup_filter()
        self._register_handlers()

    @property
    def launch_mode(self) -> LaunchMode:
        return self._launch_mode

    def states(self) -> List[State]:
        return self._states

    async def process_start(
            self,
            manager: DialogManager,
            start_data: Any,
            state: Optional[State] = None,
    ) -> None:
        if state is None:
            state = self._states[0]
        logger.debug("Dialog start: %s (%s)", state, self)
        await manager.switch_to(state)
        await self._process_callback(self.on_start, start_data, manager)

    async def _process_callback(
            self, callback: Optional[OnDialogEvent], *args, **kwargs,
    ):
        if callback:
            await callback(*args, **kwargs)

    async def _current_window(
            self, manager: DialogManager,
    ) -> WindowProtocol:
        try:
            return self.windows[manager.current_context().state]
        except KeyError as e:
            raise UnregisteredWindowError(
                f"No window found for `{manager.current_context().state}` "
                f"Current state group is `{self.states_group_name()}`",
            ) from e

    async def load_data(
            self, manager: DialogManager,
    ) -> Dict:
        data = await manager.load_data()
        data.update(await self.getter(**manager.middleware_data))
        return data

    async def render(self, manager: DialogManager) -> NewMessage:
        logger.debug("Dialog render (%s)", self)
        window = await self._current_window(manager)
        new_message = await window.render(self, manager)
        add_indent_id(new_message, manager.current_context().id)
        return new_message

    async def _message_handler(
            self, message: Message, dialog_manager: DialogManager,
    ):
        text, data = split_reply_callback(message.text)
        old_context = dialog_manager.current_context()
        window = await self._current_window(dialog_manager)

        if data:
            query = CallbackQuery(
                id="0",
                message=message,
                data=data,
                from_user=message.from_user,
                chat_instance=message.chat.username,
            )
            await window.process_callback(query, self, dialog_manager)
        else:
            await window.process_message(message, self, dialog_manager)
        if dialog_manager.has_context():
            if dialog_manager.current_context() == old_context:  # same dialog
                await dialog_manager.show()

    async def _callback_handler(
            self,
            callback: CallbackQuery,
            dialog_manager: DialogManager,
    ):
        old_context = dialog_manager.current_context()
        intent_id, callback_data = remove_indent_id(callback.data)
        cleaned_callback = callback.model_copy(update={"data": callback_data})
        window = await self._current_window(dialog_manager)
        await window.process_callback(cleaned_callback, self, dialog_manager)
        if dialog_manager.has_context():
            if dialog_manager.current_context() == old_context:  # same dialog
                await dialog_manager.show()
        await dialog_manager.answer_callback()

    def _setup_filter(self):
        intent_filter = IntentFilter(
            aiogd_intent_state_group=self.states_group(),
        )
        for observer in self.observers.values():
            observer.filter(intent_filter)

    def _register_handlers(self) -> None:
        self.callback_query.register(self._callback_handler)
        self.message.register(self._message_handler)

    def states_group(self) -> Type[StatesGroup]:
        return self._states_group

    def states_group_name(self) -> str:
        return self._states_group.__full_group_name__

    async def process_result(
            self,
            start_data: Data,
            result: Any,
            manager: DialogManager,
    ) -> None:
        await self._process_callback(
            self.on_process_result, start_data, result, manager,
        )

    def include_router(self, router: Router) -> Router:
        raise TypeError("Dialog cannot include routers")

    async def process_close(
            self, result: Any, manager: DialogManager,
    ) -> None:
        await self._process_callback(self.on_close, result, manager)

    def find(self, widget_id) -> Optional[W]:
        for w in self.windows.values():
            widget = w.find(widget_id)
            if widget:
                return widget
        return None

    def __repr__(self):
        return f"<{self.__class__.__qualname__}({self.states_group()})>"



from typing import Dict

from aiogram import Router

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.internal import DialogManagerFactory
from aiogram_dialog.api.protocols import (
    DialogManager, DialogRegistryProtocol,
    MediaIdStorageProtocol, MessageManagerProtocol,
)
from .manager import ManagerImpl


class DefaultManagerFactory(DialogManagerFactory):
    def __init__(
            self,
            message_manager: MessageManagerProtocol,
            media_id_storage: MediaIdStorageProtocol,
    ) -> None:
        self.message_manager = message_manager
        self.media_id_storage = media_id_storage

    def __call__(
            self, event: ChatEvent, data: Dict,
            registry: DialogRegistryProtocol,
            router: Router,
    ) -> DialogManager:
        return ManagerImpl(
            event=event,
            data=data,
            message_manager=self.message_manager,
            media_id_storage=self.media_id_storage,
            registry=registry,
            router=router,
        )
